from typing import Optional, AsyncIterator

from .OlvidClient import OlvidClient
from .. import datatypes
from ..internal import admin

from .logger import command_logger


class OlvidAdminClient(OlvidClient):
	"""
	OlvidAdminClient: OlvidClient extended to add access to gRPC admin methods.

	As OlvidAdminClient extends OlvidClient it has the same constraints, for example it needs to find a valid client_key to
	connect to a daemon (see OlvidClient for more information).
	An admin client needs to use an admin_client_key (a key not associated to an identity, with an identity_id equal to 0).
	It uses same mechanism as OlvidBot to look for an admin client key except that the key file is named *.admin_client_key*
	and environment variable name is OLVID_ADMIN_CLIENT_KEY.

	As in OlvidClient OlvidAdminClient implements gRPC methods as python methods.
	You can find methods using the same name as in gRPC but using snake case.

	If an admin client wants to use normal command and notification api you will need to specify the identity you want to use.
	Use the current_identity_id property to specify the identity you want to use. It will persist for any future api call.
	"""
	_KEY_ENV_VARIABLE_NAME = "OLVID_ADMIN_CLIENT_KEY"
	_KEY_FILE_PATH = ".admin_client_key"

	def __init__(self, identity_id: int, client_key: Optional[str] = None, server_target: Optional[str] = None, parent_client: Optional['OlvidClient'] = None):
		try:
			super().__init__(client_key, server_target, parent_client)
		except ValueError:
			raise ValueError("Admin client key not found")

		# overwrite parent client type
		self._parent_client: Optional["OlvidAdminClient"] = parent_client

		# admin client need to specify an identity id in all requests metadata
		self._current_identity_id: int = identity_id

		# generate admin stubs
		self._stubs.create_admin_stubs()

	#####
	# current identity property
	#####
	@property
	def current_identity_id(self) -> int:
		return self._current_identity_id

	@current_identity_id.setter
	def current_identity_id(self, identity_id: int) -> None:
		self._current_identity_id = identity_id

	async def get_current_identity_details(self) -> Optional[datatypes.IdentityDetails]:
		return (await self.admin_identity_admin_get(identity_id=self._current_identity_id)).details

	#####
	# override grpc metadata property, to add identity id
	#####
	@property
	def grpc_metadata(self) -> list[tuple[str, str]]:
		return [
			("daemon-client-key", self._client_key),
			("daemon-identity-id", str(self.current_identity_id))
		]

	####################################################################################################################
	##### WARNING: DO NOT EDIT: this code is automatically generated, see overlay_generator/generate_olvid_client_code.py
	####################################################################################################################
	# ClientKeyAdminService
	def admin_client_key_list(self, filter: datatypes.ClientKeyFilter = None) -> AsyncIterator[datatypes.ClientKey]:
		command_logger.info(f'{self.__class__.__name__}: command: ClientKeyList')
	
		async def iterator(message_iterator: AsyncIterator[admin.ClientKeyListResponse]) -> AsyncIterator[datatypes.ClientKey]:
			async for message in message_iterator:
				for element in message.client_keys:
					yield element
		return iterator(self._stubs.clientKeyAdminStub.client_key_list(admin.ClientKeyListRequest(client=self, filter=filter)))
	
	async def admin_client_key_get(self, client_key: str) -> datatypes.ClientKey:
		command_logger.info(f'{self.__class__.__name__}: command: ClientKeyGet')
		response: admin.ClientKeyGetResponse = await self._stubs.clientKeyAdminStub.client_key_get(admin.ClientKeyGetRequest(client=self, client_key=client_key))
		return response.client_key
	
	async def admin_client_key_new(self, name: str, identity_id: int) -> datatypes.ClientKey:
		command_logger.info(f'{self.__class__.__name__}: command: ClientKeyNew')
		response: admin.ClientKeyNewResponse = await self._stubs.clientKeyAdminStub.client_key_new(admin.ClientKeyNewRequest(client=self, name=name, identity_id=identity_id))
		return response.client_key
	
	async def admin_client_key_delete(self, client_key: str) -> None:
		command_logger.info(f'{self.__class__.__name__}: command: ClientKeyDelete')
		await self._stubs.clientKeyAdminStub.client_key_delete(admin.ClientKeyDeleteRequest(client=self, client_key=client_key))
	
	# IdentityAdminService
	def admin_identity_list(self, filter: datatypes.IdentityFilter = None) -> AsyncIterator[datatypes.Identity]:
		command_logger.info(f'{self.__class__.__name__}: command: IdentityList')
	
		async def iterator(message_iterator: AsyncIterator[admin.IdentityListResponse]) -> AsyncIterator[datatypes.Identity]:
			async for message in message_iterator:
				for element in message.identities:
					yield element
		return iterator(self._stubs.identityAdminStub.identity_list(admin.IdentityListRequest(client=self, filter=filter)))
	
	async def admin_identity_admin_get(self, identity_id: int) -> datatypes.Identity:
		command_logger.info(f'{self.__class__.__name__}: command: IdentityAdminGet')
		response: admin.IdentityAdminGetResponse = await self._stubs.identityAdminStub.identity_admin_get(admin.IdentityAdminGetRequest(client=self, identity_id=identity_id))
		return response.identity
	
	async def admin_identity_delete(self, identity_id: int) -> None:
		command_logger.info(f'{self.__class__.__name__}: command: IdentityDelete')
		await self._stubs.identityAdminStub.identity_delete(admin.IdentityDeleteRequest(client=self, identity_id=identity_id))
	
	async def admin_identity_new(self, identity_details: datatypes.IdentityDetails, server_url: str = "", api_key: str = "") -> datatypes.Identity:
		command_logger.info(f'{self.__class__.__name__}: command: IdentityNew')
		response: admin.IdentityNewResponse = await self._stubs.identityAdminStub.identity_new(admin.IdentityNewRequest(client=self, identity_details=identity_details, server_url=server_url, api_key=api_key))
		return response.identity
	
	async def admin_identity_keycloak_new(self, configuration_link: str) -> datatypes.Identity:
		command_logger.info(f'{self.__class__.__name__}: command: IdentityKeycloakNew')
		response: admin.IdentityKeycloakNewResponse = await self._stubs.identityAdminStub.identity_keycloak_new(admin.IdentityKeycloakNewRequest(client=self, configuration_link=configuration_link))
		return response.identity
	
	


import copy
import logging

import grpc

from ClientHolder import ClientHolder, ClientWrapper
from olvid import OlvidClient, datatypes, listeners
from utils import send_message_wait_and_check_content


# noinspection PyProtectedMember
async def test_check_list_and_get_content(client_holder: ClientHolder, client1: ClientWrapper, client2: ClientWrapper):
	# check number of contacts is correct
	identity_count = len(client_holder.clients)
	contacts1 = [c async for c in client1.contact_list()]
	assert len(contacts1) == identity_count - 1
	contacts2 = [c async for c in client2.contact_list()]
	assert len(contacts2) == identity_count - 1

	# check contact_lis and contact_get returns same data
	[(await client1.contact_get(contact_id=c.id))._test_assertion(c) for c in contacts1]
	[(await client1.contact_get(contact_id=c.id))._test_assertion(c) for c in contacts1]


# noinspection PyProtectedMember
async def test_down_upgrade_one_to_one_and_locked_discussions(client1: ClientWrapper, client2: ClientWrapper):
	# get other client contact on each side
	contact1: datatypes.Contact = await client1.get_contact_associated_to_another_client(client2)
	contact2: datatypes.Contact = await client2.get_contact_associated_to_another_client(client1)

	if not contact1.has_one_to_one_discussion or not contact2.has_one_to_one_discussion:
		logging.warning(f"test_contact: contacts are not one to one, upgrading channel first ({client1.identity.id} <-> {client2.identity.id})")
		await upgrade_contact_to_one_to_one(client1, client2, contact1, contact2, ideal_discussion1=datatypes.Discussion(title=contact1.display_name), ideal_discussion2=datatypes.Discussion(title=contact2.display_name))

	# get other client discussion on each side
	discussion1: datatypes.Discussion = await client1.get_discussion_associated_to_another_client(client2)
	discussion2: datatypes.Discussion = await client2.get_discussion_associated_to_another_client(client1)

	# send message in discussions to be sure it's not empty
	await send_message_wait_and_check_content(client1, client2, body="Do not let discussion empty")

	# mark as not one to one cause we will use this value to check notif content
	contact1.has_one_to_one_discussion = False
	contact2.has_one_to_one_discussion = False

	# create bot to check contact deletion
	bots: list[OlvidClient] = [
		# contact deleted on each side
		client1.create_notification_bot(listeners.ContactDeletedListener(handler=client1.get_check_content_handler(contact1, notification_type=listeners.NOTIFICATIONS.CONTACT_DELETED), count=1)),
		client2.create_notification_bot(listeners.ContactDeletedListener(handler=client2.get_check_content_handler(contact2, notification_type=listeners.NOTIFICATIONS.CONTACT_DELETED), count=1)),
		# discussion locked on each side
		client1.create_notification_bot(listeners.DiscussionLockedListener(handler=client1.get_check_content_handler(discussion1, notification_type=listeners.NOTIFICATIONS.DISCUSSION_LOCKED), count=1)),
		client2.create_notification_bot(listeners.DiscussionLockedListener(handler=client2.get_check_content_handler(discussion2, notification_type=listeners.NOTIFICATIONS.DISCUSSION_LOCKED), count=1)),
	]

	# delete contact (downgrade to not one to one discussion)
	await client1.contact_downgrade_one_to_one_discussion(contact_id=contact1.id)

	# wait for notifications
	for bot in bots:
		await bot.wait_for_listeners_end()
		await bot.stop()

	# check contacts are still here and have been downgraded
	contact1._test_assertion(await client1.contact_get(contact1.id))
	contact2._test_assertion(await client2.contact_get(contact2.id))

	# check discussion appear as locked
	assert len([ld async for ld in client1.discussion_locked_list() if ld.id == discussion1.id]) == 1, "locked discussion cannot be listed"
	assert len([ld async for ld in client2.discussion_locked_list() if ld.id == discussion2.id]) == 1, "locked discussion cannot be listed"

	# check discussion is not listed anymore
	assert len([ld async for ld in client1.discussion_list() if ld.id == discussion1.id]) == 0, "locked discussion still listed"
	assert len([ld async for ld in client2.discussion_list() if ld.id == discussion2.id]) == 0, "locked discussion still listed"

	# check discussion is not getable
	try:
		await client1.discussion_get(discussion_id=discussion1.id)
		assert False, "discussion still accessible"
	except grpc.aio.AioRpcError:
		pass
	try:
		await client2.discussion_get(discussion_id=discussion2.id)
		assert False, "discussion still accessible"
	except grpc.aio.AioRpcError:
		pass

	# actually we can do it, that's probably what we want to do
	# check we cannot list messages in a locked discussion
	# try:
	# 	messages = [m async for m in client1.message_list(datatypes.MessageFilter(discussion_id=discussion1.id))]
	# 	assert False, f"can list messages in a locked discussion: {messages}"
	# except grpc.aio.AioRpcError:
	# 	pass
	# try:
	# 	messages = [m async for m in client2.message_list(datatypes.MessageFilter(discussion_id=discussion2.id))]
	# 	assert False, f"can list messages in a locked discussion: {messages}"
	# except grpc.aio.AioRpcError:
	# 	pass

	# check we cannot send a message in a locked discussion
	try:
		await client1.message_send(discussion_id=discussion1.id, body="This message MUST NOT be sent")
		assert False, "can post in a locked discussion"
	except grpc.aio.AioRpcError:
		pass
	try:
		await client2.message_send(discussion_id=discussion2.id, body="This message MUST NOT be sent")
		assert False, "can post in a locked discussion"
	except grpc.aio.AioRpcError:
		pass

	# delete locked discussion
	await client1.discussion_locked_delete(discussion_id=discussion1.id)
	await client2.discussion_locked_delete(discussion_id=discussion2.id)

	# check discussion locked had been deleted
	assert len([ld async for ld in client1.discussion_locked_list() if ld.id == discussion1.id]) == 0, "deleted locked discussion can be listed"
	assert len([ld async for ld in client2.discussion_locked_list() if ld.id == discussion2.id]) == 0, "deleted locked discussion can be listed"

	# check discussion is not listed anymore
	assert len([ld async for ld in client1.discussion_list() if ld.id == discussion1.id]) == 0, "locked discussion still listed as normal after deletion"
	assert len([ld async for ld in client2.discussion_list() if ld.id == discussion2.id]) == 0, "locked discussion still listed as normal after deletion"

	# check discussion is not getable
	try:
		await client1.discussion_get(discussion_id=discussion1.id)
		assert False, "discussion still accessible"
	except grpc.aio.AioRpcError:
		pass
	try:
		await client2.discussion_get(discussion_id=discussion2.id)
		assert False, "discussion still accessible"
	except grpc.aio.AioRpcError:
		pass

	await upgrade_contact_to_one_to_one(client1, client2, contact1, contact2,
										ideal_discussion1=datatypes.Discussion(title=contact1.display_name),
										ideal_discussion2=datatypes.Discussion(title=contact2.display_name))

	new_contact_1: datatypes.Contact = await client1.contact_get(contact_id=contact1.id)
	# get new contact and check content
	new_contact_1._test_assertion(contact1)
	new_contact_2: datatypes.Contact = await client2.contact_get(contact_id=contact2.id)
	new_contact_2._test_assertion(contact2)


# noinspection PyProtectedMember
async def upgrade_contact_to_one_to_one(client1: ClientWrapper, client2: ClientWrapper, contact1: datatypes.Contact, contact2: datatypes.Contact, ideal_discussion1: datatypes.Discussion, ideal_discussion2: datatypes.Discussion):
	# prepare notification bots for one to one upgrade process
	contact1.has_one_to_one_discussion = True
	contact2.has_one_to_one_discussion = True

	# noinspection PyProtectedMember
	async def accept_invitation_handler(invitation: datatypes.Invitation):
		await invitation._client.invitation_accept(invitation.id)
	bots = [
		# client1: invitation sent, invitation deleted, contact new, discussion new
		client1.create_notification_bot(listeners.InvitationSentListener(handler=client1.get_check_content_handler(datatypes.Invitation(status=datatypes.Invitation.Status.STATUS_ONE_TO_ONE_INVITATION_WAIT_IT_TO_ACCEPT, display_name=contact1.display_name, sas=""), notification_type=listeners.NOTIFICATIONS.INVITATION_SENT), count=1)),
		client1.create_notification_bot(listeners.InvitationDeletedListener(handler=client1.get_check_content_handler(datatypes.Invitation(status=datatypes.Invitation.Status.STATUS_ONE_TO_ONE_INVITATION_WAIT_IT_TO_ACCEPT, display_name=contact1.display_name, sas=""), notification_type=listeners.NOTIFICATIONS.INVITATION_DELETED), count=1)),
		client1.create_notification_bot(listeners.ContactNewListener(handler=client1.get_check_content_handler(contact1, notification_type=listeners.NOTIFICATIONS.CONTACT_NEW), count=1)),
		client1.create_notification_bot(listeners.DiscussionNewListener(handler=client1.get_check_content_handler(ideal_discussion1, notification_type=listeners.NOTIFICATIONS.DISCUSSION_NEW), count=1)),

		# client2: invitation received (check content and accept it), invitation deleted, contact new, discussion new
		client2.create_notification_bot(listeners.InvitationReceivedListener(handler=client2.get_check_content_handler(datatypes.Invitation(status=datatypes.Invitation.Status.STATUS_ONE_TO_ONE_INVITATION_WAIT_YOU_TO_ACCEPT, display_name=contact2.display_name, sas=""), notification_type=listeners.NOTIFICATIONS.INVITATION_SENT), count=1)),
		client2.create_notification_bot(listeners.InvitationReceivedListener(handler=accept_invitation_handler, count=1)),
		client2.create_notification_bot(listeners.InvitationDeletedListener(handler=client2.get_check_content_handler(datatypes.Invitation(status=datatypes.Invitation.Status.STATUS_ONE_TO_ONE_INVITATION_WAIT_YOU_TO_ACCEPT, display_name=contact2.display_name, sas=""), notification_type=listeners.NOTIFICATIONS.INVITATION_DELETED), count=1)),
		client2.create_notification_bot(listeners.ContactNewListener(handler=client2.get_check_content_handler(contact2, notification_type=listeners.NOTIFICATIONS.CONTACT_NEW), count=1)),
		client2.create_notification_bot(listeners.DiscussionNewListener(handler=client2.get_check_content_handler(ideal_discussion2, notification_type=listeners.NOTIFICATIONS.DISCUSSION_NEW), count=1)),
	]

	await client1.contact_invite_to_one_to_one_discussion(contact_id=contact1.id)

	for bot in bots:
		await bot.wait_for_listeners_end()
		await bot.stop()

	# wait for channel to be established
	await client1.wait_for_channel_creation()
	await client2.wait_for_channel_creation()
	logging.debug(f"channel created {client1.identity.id} <-> {client2.identity.id}")

	# get new contact and check content
	new_contact_1: datatypes.Contact = await client1.contact_get(contact_id=contact1.id)
	new_contact_1._test_assertion(contact1)
	new_contact_2: datatypes.Contact = await client2.contact_get(contact_id=contact2.id)
	new_contact_2._test_assertion(contact2)


async def test_update_details(client: ClientWrapper, client_holder: ClientHolder):
	# prepare new details for client2
	prev_details: datatypes.IdentityDetails = client.identity.details
	new_details: datatypes.IdentityDetails = datatypes.IdentityDetails(first_name=prev_details.first_name + "-UPDATED", last_name=prev_details.last_name + "-UPDATED", position=prev_details.position + "-UPDATED", company=prev_details.company + "-UPDATED")

	try:
		await update_details(client, client_holder, new_details)
		# update store identity in clientWrapper, we need up to date details to find associated contact / discussion in other clients
		client.identity = await client.identity_get()

		await update_details(client, client_holder, prev_details)
		# update store identity in clientWrapper, we need up to date details to find associated contact / discussion in other clients
		client.identity = await client.identity_get()

	# if something goes wrong try to restore previous details before exiting
	except Exception:
		await client.identity_update_details(prev_details)
		raise


# noinspection PyProtectedMember
async def update_details(client: ClientWrapper, client_holder: ClientHolder, new_details: datatypes.IdentityDetails):
	# prepare bot for notifications
	bots: list[OlvidClient] = []

	for other_client in [c for c in client_holder.clients if c != client]:
		contact: datatypes.Contact = await other_client.get_contact_associated_to_another_client(client)
		previous_details = contact.details
		contact.display_name = f"{new_details.first_name} {new_details.last_name} ({new_details.position} @ {new_details.company})"
		contact.details = new_details
		discussion: datatypes.Discussion = await other_client.get_discussion_associated_to_another_client(client)
		previous_title: str = discussion.title
		discussion.title = contact.display_name
		bots.extend([
			other_client.create_notification_bot(listeners.ContactDetailsUpdatedListener(handler=other_client.get_check_content_handler(contact._clone(), previous_details._clone(), notification_type=listeners.NOTIFICATIONS.CONTACT_DETAILS_UPDATED), count=1)),
			other_client.create_notification_bot(listeners.DiscussionTitleUpdatedListener(handler=other_client.get_check_content_handler(discussion._clone(), copy.deepcopy(previous_title), notification_type=listeners.NOTIFICATIONS.DISCUSSION_TITLE_UPDATED), count=1)),
		])

	await client.identity_update_details(new_details=new_details)

	for bot in bots:
		await bot.wait_for_listeners_end()
		await bot.stop()


async def test_contact(client_holder: ClientHolder):
	for c1, c2 in client_holder.get_all_client_pairs():
		logging.info(f"{c1.identity.id} <-> {c2.identity.id}: prepare tests")
		await test_check_list_and_get_content(client_holder, c1, c2)
		logging.info(f"{c1.identity.id} <-> {c2.identity.id}: downgrade one to one")
		await test_down_upgrade_one_to_one_and_locked_discussions(c1, c2)
		logging.info(f"{c1.identity.id} <-> {c2.identity.id}: update contact details")
		await test_update_details(c1, client_holder)



import datetime
from typing import Optional

from ..core.OlvidBot import OlvidBot
from ..listeners.ListenersImplementation import MessageReceivedListener, MessageSentListener
from .. import datatypes


class DiscussionRetentionPolicyBot(OlvidBot):
	def __init__(self, retention_delay_s: int = None, discussion_retention_number: int = None, global_retention_number: int = None):
		super().__init__()

		self.retention_delay_s: Optional[int] = retention_delay_s
		self.discussion_retention_number: Optional[int] = discussion_retention_number
		self.global_retention_number: Optional[int] = global_retention_number

		self.add_listener(listener=MessageReceivedListener(handler=self.message_received_handler))
		self.add_listener(listener=MessageSentListener(handler=self.message_sent_handler))

		# on start clean everything
		async def start_task():
			await self._clean_globally()
			async for d in self.discussion_list():
				await self._clean_discussion_messages(d.id)

		self.add_background_task(start_task())

	# delete every received messages
	async def message_received_handler(self, message: datatypes.Message):
		if self.retention_delay_s or self.discussion_retention_number:
			await self._clean_discussion_messages(message.discussion_id)
		if self.global_retention_number:
			await self._clean_globally()

	async def message_sent_handler(self, message: datatypes.Message):
		if self.retention_delay_s or self.discussion_retention_number:
			await self._clean_discussion_messages(message.discussion_id)
		if self.global_retention_number:
			await self._clean_globally()

	async def _clean_discussion_messages(self, discussion_id: int):
		discussion_messages: list[datatypes.Message] = [m async for m in self.message_list(filter=datatypes.MessageFilter(discussion_id=discussion_id))]
		messages_to_delete: set[datatypes.Message] = set()

		if self.retention_delay_s is not None:
			now_ms: int = round(datetime.datetime.now().timestamp() * 1000)
			for message in discussion_messages:
				if message.timestamp < now_ms - self.retention_delay_s * 1000:
					messages_to_delete.add(message)

		if self.discussion_retention_number is not None:
			if len(discussion_messages) > self.discussion_retention_number:
				messages_to_delete.update(discussion_messages[:-self.discussion_retention_number])

		for message_to_delete in messages_to_delete:
			await self.message_delete(message_id=message_to_delete.id)

	async def _clean_globally(self):
		all_messages: list[datatypes.Message] = [m async for m in self.message_list()]

		if self.global_retention_number:
			if len(all_messages) > self.global_retention_number:
				for message_to_delete in all_messages[:-self.global_retention_number]:
					await self.message_delete(message_id=message_to_delete.id)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: olvid/daemon/datatypes/v1/group.proto
# Protobuf Python Version: 5.27.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    3,
    '',
    'olvid/daemon/datatypes/v1/group.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%olvid/daemon/datatypes/v1/group.proto\x12\x19olvid.daemon.datatypes.v1\"\x93\x08\n\x05Group\x12\x0e\n\x02id\x18\x01 \x01(\x04R\x02id\x12\x39\n\x04type\x18\x02 \x01(\x0e\x32%.olvid.daemon.datatypes.v1.Group.TypeR\x04type\x12r\n\x16\x61\x64vanced_configuration\x18\x03 \x01(\x0b\x32\x36.olvid.daemon.datatypes.v1.Group.AdvancedConfigurationH\x00R\x15\x61\x64vancedConfiguration\x88\x01\x01\x12Z\n\x0fown_permissions\x18\x04 \x01(\x0b\x32\x31.olvid.daemon.datatypes.v1.GroupMemberPermissionsR\x0eownPermissions\x12@\n\x07members\x18\x05 \x03(\x0b\x32&.olvid.daemon.datatypes.v1.GroupMemberR\x07members\x12V\n\x0fpending_members\x18\x06 \x03(\x0b\x32-.olvid.daemon.datatypes.v1.PendingGroupMemberR\x0ependingMembers\x12,\n\x12update_in_progress\x18\x07 \x01(\x08R\x10updateInProgress\x12)\n\x10keycloak_managed\x18\x08 \x01(\x08R\x0fkeycloakManaged\x12\x12\n\x04name\x18\t \x01(\tR\x04name\x12 \n\x0b\x64\x65scription\x18\n \x01(\tR\x0b\x64\x65scription\x12\x1e\n\x0bhas_a_photo\x18\x0b \x01(\x08R\thasAPhoto\x1a\x9d\x02\n\x15\x41\x64vancedConfiguration\x12\x1b\n\tread_only\x18\x01 \x01(\x08R\x08readOnly\x12h\n\rremote_delete\x18\x02 \x01(\x0e\x32\x43.olvid.daemon.datatypes.v1.Group.AdvancedConfiguration.RemoteDeleteR\x0cremoteDelete\"}\n\x0cRemoteDelete\x12\x1d\n\x19REMOTE_DELETE_UNSPECIFIED\x10\x00\x12\x18\n\x14REMOTE_DELETE_NOBODY\x10\x01\x12\x18\n\x14REMOTE_DELETE_ADMINS\x10\x02\x12\x1a\n\x16REMOTE_DELETE_EVERYONE\x10\x03\"k\n\x04Type\x12\x14\n\x10TYPE_UNSPECIFIED\x10\x00\x12\x11\n\rTYPE_STANDARD\x10\x01\x12\x13\n\x0fTYPE_CONTROLLED\x10\x02\x12\x12\n\x0eTYPE_READ_ONLY\x10\x03\x12\x11\n\rTYPE_ADVANCED\x10\x04\x42\x19\n\x17_advanced_configuration\"\x81\x01\n\x0bGroupMember\x12\x1d\n\ncontact_id\x18\x01 \x01(\x04R\tcontactId\x12S\n\x0bpermissions\x18\x02 \x01(\x0b\x32\x31.olvid.daemon.datatypes.v1.GroupMemberPermissionsR\x0bpermissions\"\xf3\x01\n\x12PendingGroupMember\x12*\n\x11pending_member_id\x18\x01 \x01(\x04R\x0fpendingMemberId\x12\x1d\n\ncontact_id\x18\x02 \x01(\x04R\tcontactId\x12!\n\x0c\x64isplay_name\x18\x03 \x01(\tR\x0b\x64isplayName\x12\x1a\n\x08\x64\x65\x63lined\x18\x04 \x01(\x08R\x08\x64\x65\x63lined\x12S\n\x0bpermissions\x18\x05 \x01(\x0b\x32\x31.olvid.daemon.datatypes.v1.GroupMemberPermissionsR\x0bpermissions\"\xfb\x01\n\x16GroupMemberPermissions\x12\x14\n\x05\x61\x64min\x18\x01 \x01(\x08R\x05\x61\x64min\x12\x34\n\x16remote_delete_anything\x18\x02 \x01(\x08R\x14remoteDeleteAnything\x12I\n\"edit_or_remote_delete_own_messages\x18\x03 \x01(\x08R\x1d\x65\x64itOrRemoteDeleteOwnMessages\x12\'\n\x0f\x63hange_settings\x18\x04 \x01(\x08R\x0e\x63hangeSettings\x12!\n\x0csend_message\x18\x05 \x01(\x08R\x0bsendMessage\"\xf2\x07\n\x0bGroupFilter\x12>\n\x04type\x18\x01 \x01(\x0e\x32%.olvid.daemon.datatypes.v1.Group.TypeH\x00R\x04type\x88\x01\x01\x12G\n\x05\x65mpty\x18\x02 \x01(\x0e\x32,.olvid.daemon.datatypes.v1.GroupFilter.EmptyH\x01R\x05\x65mpty\x88\x01\x01\x12G\n\x05photo\x18\x03 \x01(\x0e\x32,.olvid.daemon.datatypes.v1.GroupFilter.PhotoH\x02R\x05photo\x88\x01\x01\x12P\n\x08keycloak\x18\x04 \x01(\x0e\x32/.olvid.daemon.datatypes.v1.GroupFilter.KeycloakH\x03R\x08keycloak\x88\x01\x01\x12k\n\x16own_permissions_filter\x18\x05 \x01(\x0b\x32\x30.olvid.daemon.datatypes.v1.GroupPermissionFilterH\x04R\x14ownPermissionsFilter\x88\x01\x01\x12$\n\x0bname_search\x18\x06 \x01(\tH\x05R\nnameSearch\x88\x01\x01\x12\x32\n\x12\x64\x65scription_search\x18\x07 \x01(\tH\x06R\x11\x64\x65scriptionSearch\x88\x01\x01\x12S\n\x0emember_filters\x18\x08 \x03(\x0b\x32,.olvid.daemon.datatypes.v1.GroupMemberFilterR\rmemberFilters\x12i\n\x16pending_member_filters\x18\t \x03(\x0b\x32\x33.olvid.daemon.datatypes.v1.PendingGroupMemberFilterR\x14pendingMemberFilters\">\n\x05\x45mpty\x12\x15\n\x11\x45MPTY_UNSPECIFIED\x10\x00\x12\x0c\n\x08\x45MPTY_IS\x10\x02\x12\x10\n\x0c\x45MPTY_IS_NOT\x10\x01\"J\n\x08Keycloak\x12\x18\n\x14KEYCLOAK_UNSPECIFIED\x10\x00\x12\x0f\n\x0bKEYCLOAK_IS\x10\x02\x12\x13\n\x0fKEYCLOAK_IS_NOT\x10\x01\"@\n\x05Photo\x12\x15\n\x11PHOTO_UNSPECIFIED\x10\x00\x12\r\n\tPHOTO_HAS\x10\x02\x12\x11\n\rPHOTO_HAS_NOT\x10\x01\x42\x07\n\x05_typeB\x08\n\x06_emptyB\x08\n\x06_photoB\x0b\n\t_keycloakB\x19\n\x17_own_permissions_filterB\x0e\n\x0c_name_searchB\x15\n\x13_description_search\"\xaf\x01\n\x11GroupMemberFilter\x12\"\n\ncontact_id\x18\x01 \x01(\x04H\x00R\tcontactId\x88\x01\x01\x12W\n\x0bpermissions\x18\x05 \x01(\x0b\x32\x30.olvid.daemon.datatypes.v1.GroupPermissionFilterH\x01R\x0bpermissions\x88\x01\x01\x42\r\n\x0b_contact_idB\x0e\n\x0c_permissions\"\x80\x05\n\x18PendingGroupMemberFilter\x12_\n\nis_contact\x18\x01 \x01(\x0e\x32;.olvid.daemon.datatypes.v1.PendingGroupMemberFilter.ContactH\x00R\tisContact\x88\x01\x01\x12\x64\n\x0chas_declined\x18\x02 \x01(\x0e\x32<.olvid.daemon.datatypes.v1.PendingGroupMemberFilter.DeclinedH\x01R\x0bhasDeclined\x88\x01\x01\x12\"\n\ncontact_id\x18\x03 \x01(\x04H\x02R\tcontactId\x88\x01\x01\x12\x33\n\x13\x64isplay_name_search\x18\x04 \x01(\tH\x03R\x11\x64isplayNameSearch\x88\x01\x01\x12W\n\x0bpermissions\x18\x05 \x01(\x0b\x32\x30.olvid.daemon.datatypes.v1.GroupPermissionFilterH\x04R\x0bpermissions\x88\x01\x01\"F\n\x07\x43ontact\x12\x17\n\x13\x43ONTACT_UNSPECIFIED\x10\x00\x12\x0e\n\nCONTACT_IS\x10\x01\x12\x12\n\x0e\x43ONTACT_IS_NOT\x10\x02\"L\n\x08\x44\x65\x63lined\x12\x18\n\x14\x44\x45\x43LINED_UNSPECIFIED\x10\x00\x12\x10\n\x0c\x44\x45\x43LINED_HAS\x10\x01\x12\x14\n\x10\x44\x45\x43LINED_HAS_NOT\x10\x02\x42\r\n\x0b_is_contactB\x0f\n\r_has_declinedB\r\n\x0b_contact_idB\x16\n\x14_display_name_searchB\x0e\n\x0c_permissions\"\x88\n\n\x15GroupPermissionFilter\x12Q\n\x05\x61\x64min\x18\x01 \x01(\x0e\x32\x36.olvid.daemon.datatypes.v1.GroupPermissionFilter.AdminH\x00R\x05\x61\x64min\x88\x01\x01\x12\x64\n\x0csend_message\x18\x02 \x01(\x0e\x32<.olvid.daemon.datatypes.v1.GroupPermissionFilter.SendMessageH\x01R\x0bsendMessage\x88\x01\x01\x12\x80\x01\n\x16remote_delete_anything\x18\x03 \x01(\x0e\x32\x45.olvid.daemon.datatypes.v1.GroupPermissionFilter.RemoteDeleteAnythingH\x02R\x14remoteDeleteAnything\x88\x01\x01\x12\x9d\x01\n\"edit_or_remote_delete_own_messages\x18\x04 \x01(\x0e\x32M.olvid.daemon.datatypes.v1.GroupPermissionFilter.EditOrRemoteDeleteOwnMessageH\x03R\x1d\x65\x64itOrRemoteDeleteOwnMessages\x88\x01\x01\x12m\n\x0f\x63hange_settings\x18\x05 \x01(\x0e\x32?.olvid.daemon.datatypes.v1.GroupPermissionFilter.ChangeSettingsH\x04R\x0e\x63hangeSettings\x88\x01\x01\">\n\x05\x41\x64min\x12\x15\n\x11\x41\x44MIN_UNSPECIFIED\x10\x00\x12\x0c\n\x08\x41\x44MIN_IS\x10\x01\x12\x10\n\x0c\x41\x44MIN_IS_NOT\x10\x02\"Z\n\x0bSendMessage\x12\x1c\n\x18SEND_MESSAGE_UNSPECIFIED\x10\x00\x12\x14\n\x10SEND_MESSAGE_CAN\x10\x01\x12\x17\n\x13SEND_MESSAGE_CANNOT\x10\x02\"\x81\x01\n\x14RemoteDeleteAnything\x12&\n\"REMOTE_DELETE_ANYTHING_UNSPECIFIED\x10\x00\x12\x1e\n\x1aREMOTE_DELETE_ANYTHING_CAN\x10\x01\x12!\n\x1dREMOTE_DELETE_ANYTHING_CANNOT\x10\x02\"\xaa\x01\n\x1c\x45\x64itOrRemoteDeleteOwnMessage\x12\x31\n-EDIT_OR_REMOTE_DELETE_OWN_MESSAGE_UNSPECIFIED\x10\x00\x12)\n%EDIT_OR_REMOTE_DELETE_OWN_MESSAGE_CAN\x10\x01\x12,\n(EDIT_OR_REMOTE_DELETE_OWN_MESSAGE_CANNOT\x10\x02\"f\n\x0e\x43hangeSettings\x12\x1f\n\x1b\x43HANGE_SETTINGS_UNSPECIFIED\x10\x00\x12\x17\n\x13\x43HANGE_SETTINGS_CAN\x10\x01\x12\x1a\n\x16\x43HANGE_SETTINGS_CANNOT\x10\x02\x42\x08\n\x06_adminB\x0f\n\r_send_messageB\x19\n\x17_remote_delete_anythingB%\n#_edit_or_remote_delete_own_messagesB\x12\n\x10_change_settingsB\xc3\x01\n\x1d\x63om.olvid.daemon.datatypes.v1B\nGroupProtoP\x01Z\x0folvid.io/daemon\xa2\x02\x03ODD\xaa\x02\x19Olvid.Daemon.Datatypes.V1\xca\x02\x19Olvid\\Daemon\\Datatypes\\V1\xe2\x02%Olvid\\Daemon\\Datatypes\\V1\\GPBMetadata\xea\x02\x1cOlvid::Daemon::Datatypes::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'olvid.daemon.datatypes.v1.group_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\035com.olvid.daemon.datatypes.v1B\nGroupProtoP\001Z\017olvid.io/daemon\242\002\003ODD\252\002\031Olvid.Daemon.Datatypes.V1\312\002\031Olvid\\Daemon\\Datatypes\\V1\342\002%Olvid\\Daemon\\Datatypes\\V1\\GPBMetadata\352\002\034Olvid::Daemon::Datatypes::V1'
  _globals['_GROUP']._serialized_start=69
  _globals['_GROUP']._serialized_end=1112
  _globals['_GROUP_ADVANCEDCONFIGURATION']._serialized_start=691
  _globals['_GROUP_ADVANCEDCONFIGURATION']._serialized_end=976
  _globals['_GROUP_ADVANCEDCONFIGURATION_REMOTEDELETE']._serialized_start=851
  _globals['_GROUP_ADVANCEDCONFIGURATION_REMOTEDELETE']._serialized_end=976
  _globals['_GROUP_TYPE']._serialized_start=978
  _globals['_GROUP_TYPE']._serialized_end=1085
  _globals['_GROUPMEMBER']._serialized_start=1115
  _globals['_GROUPMEMBER']._serialized_end=1244
  _globals['_PENDINGGROUPMEMBER']._serialized_start=1247
  _globals['_PENDINGGROUPMEMBER']._serialized_end=1490
  _globals['_GROUPMEMBERPERMISSIONS']._serialized_start=1493
  _globals['_GROUPMEMBERPERMISSIONS']._serialized_end=1744
  _globals['_GROUPFILTER']._serialized_start=1747
  _globals['_GROUPFILTER']._serialized_end=2757
  _globals['_GROUPFILTER_EMPTY']._serialized_start=2445
  _globals['_GROUPFILTER_EMPTY']._serialized_end=2507
  _globals['_GROUPFILTER_KEYCLOAK']._serialized_start=2509
  _globals['_GROUPFILTER_KEYCLOAK']._serialized_end=2583
  _globals['_GROUPFILTER_PHOTO']._serialized_start=2585
  _globals['_GROUPFILTER_PHOTO']._serialized_end=2649
  _globals['_GROUPMEMBERFILTER']._serialized_start=2760
  _globals['_GROUPMEMBERFILTER']._serialized_end=2935
  _globals['_PENDINGGROUPMEMBERFILTER']._serialized_start=2938
  _globals['_PENDINGGROUPMEMBERFILTER']._serialized_end=3578
  _globals['_PENDINGGROUPMEMBERFILTER_CONTACT']._serialized_start=3343
  _globals['_PENDINGGROUPMEMBERFILTER_CONTACT']._serialized_end=3413
  _globals['_PENDINGGROUPMEMBERFILTER_DECLINED']._serialized_start=3415
  _globals['_PENDINGGROUPMEMBERFILTER_DECLINED']._serialized_end=3491
  _globals['_GROUPPERMISSIONFILTER']._serialized_start=3581
  _globals['_GROUPPERMISSIONFILTER']._serialized_end=4869
  _globals['_GROUPPERMISSIONFILTER_ADMIN']._serialized_start=4193
  _globals['_GROUPPERMISSIONFILTER_ADMIN']._serialized_end=4255
  _globals['_GROUPPERMISSIONFILTER_SENDMESSAGE']._serialized_start=4257
  _globals['_GROUPPERMISSIONFILTER_SENDMESSAGE']._serialized_end=4347
  _globals['_GROUPPERMISSIONFILTER_REMOTEDELETEANYTHING']._serialized_start=4350
  _globals['_GROUPPERMISSIONFILTER_REMOTEDELETEANYTHING']._serialized_end=4479
  _globals['_GROUPPERMISSIONFILTER_EDITORREMOTEDELETEOWNMESSAGE']._serialized_start=4482
  _globals['_GROUPPERMISSIONFILTER_EDITORREMOTEDELETEOWNMESSAGE']._serialized_end=4652
  _globals['_GROUPPERMISSIONFILTER_CHANGESETTINGS']._serialized_start=4654
  _globals['_GROUPPERMISSIONFILTER_CHANGESETTINGS']._serialized_end=4756
# @@protoc_insertion_point(module_scope)



####
# WARNING: DO NOT EDIT: this code is automatically generated, see overlay_generator/generate_protobuf_overlay
####

from ..notifications import *
from ..commands import *
from ..admin import *

from typing import Union


NotificationMessageType = Union[SubscribeToAttachmentReceivedNotification, AttachmentReceivedNotification, SubscribeToAttachmentUploadedNotification, AttachmentUploadedNotification, SubscribeToContactNewNotification, ContactNewNotification, SubscribeToContactDeletedNotification, ContactDeletedNotification, SubscribeToContactDetailsUpdatedNotification, ContactDetailsUpdatedNotification, SubscribeToDiscussionNewNotification, DiscussionNewNotification, SubscribeToDiscussionLockedNotification, DiscussionLockedNotification, SubscribeToDiscussionTitleUpdatedNotification, DiscussionTitleUpdatedNotification, SubscribeToDiscussionSettingsUpdatedNotification, DiscussionSettingsUpdatedNotification, SubscribeToGroupNewNotification, GroupNewNotification, SubscribeToGroupDeletedNotification, GroupDeletedNotification, SubscribeToGroupNameUpdatedNotification, GroupNameUpdatedNotification, SubscribeToGroupDescriptionUpdatedNotification, GroupDescriptionUpdatedNotification, SubscribeToGroupPendingMemberAddedNotification, GroupPendingMemberAddedNotification, SubscribeToGroupPendingMemberRemovedNotification, GroupPendingMemberRemovedNotification, SubscribeToGroupMemberJoinedNotification, GroupMemberJoinedNotification, SubscribeToGroupMemberLeftNotification, GroupMemberLeftNotification, SubscribeToGroupOwnPermissionsUpdatedNotification, GroupOwnPermissionsUpdatedNotification, SubscribeToGroupMemberPermissionsUpdatedNotification, GroupMemberPermissionsUpdatedNotification, SubscribeToGroupUpdateInProgressNotification, GroupUpdateInProgressNotification, SubscribeToGroupUpdateFinishedNotification, GroupUpdateFinishedNotification, SubscribeToIdentityCreatedNotification, IdentityCreatedNotification, SubscribeToIdentityDeletedNotification, IdentityDeletedNotification, SubscribeToIdentityUpdatedNotification, IdentityDetailsUpdatedNotification, SubscribeToInvitationReceivedNotification, InvitationReceivedNotification, SubscribeToInvitationSentNotification, InvitationSentNotification, SubscribeToInvitationDeletedNotification, InvitationDeletedNotification, SubscribeToInvitationUpdatedNotification, InvitationUpdatedNotification, SubscribeToMessageReceivedNotification, MessageReceivedNotification, SubscribeToMessageSentNotification, MessageSentNotification, SubscribeToMessageDeletedNotification, MessageDeletedNotification, SubscribeToMessageBodyUpdatedNotification, MessageBodyUpdatedNotification, SubscribeToMessageUploadedNotification, MessageUploadedNotification, SubscribeToMessageDeliveredNotification, MessageDeliveredNotification, SubscribeToMessageReadNotification, MessageReadNotification, SubscribeToMessageLocationReceivedNotification, MessageLocationReceivedNotification, SubscribeToMessageLocationSharingStartNotification, MessageLocationSharingStartNotification, SubscribeToMessageLocationSharingUpdateNotification, MessageLocationSharingUpdateNotification, SubscribeToMessageLocationSharingEndNotification, MessageLocationSharingEndNotification, SubscribeToMessageReactionAddedNotification, MessageReactionAddedNotification, SubscribeToMessageReactionUpdatedNotification, MessageReactionUpdatedNotification, SubscribeToMessageReactionRemovedNotification, MessageReactionRemovedNotification]
CommandRequestMessageType = Union[ClientKeyListRequest, ClientKeyGetRequest, ClientKeyNewRequest, ClientKeyDeleteRequest, IdentityListRequest, IdentityAdminGetRequest, IdentityNewRequest, IdentityDeleteRequest, IdentityKeycloakNewRequest, AttachmentListRequest, AttachmentGetRequest, AttachmentDeleteRequest, AttachmentDownloadRequest, ContactListRequest, ContactGetRequest, ContactDeleteRequest, ContactIntroductionRequest, ContactInviteToOneToOneDiscussionRequest, ContactDowngradeOneToOneDiscussionRequest, DiscussionListRequest, DiscussionGetRequest, DiscussionGetByContactRequest, DiscussionGetByGroupRequest, DiscussionEmptyRequest, DiscussionSettingsGetRequest, DiscussionSettingsSetRequest, DiscussionLockedListRequest, DiscussionLockedDeleteRequest, GroupListRequest, GroupGetRequest, GroupNewStandardGroupRequest, GroupNewControlledGroupRequest, GroupNewReadOnlyGroupRequest, GroupNewAdvancedGroupRequest, GroupDisbandRequest, GroupLeaveRequest, GroupUpdateRequest, GroupUnsetPhotoRequest, GroupSetPhotoRequest, IdentityGetRequest, IdentityUpdateDetailsRequest, IdentityRemovePhotoRequest, IdentitySetPhotoRequest, IdentityKeycloakBindRequest, IdentityKeycloakUnbindRequest, IdentitySetApiKeyRequest, IdentitySetConfigurationLinkRequest, InvitationListRequest, InvitationGetRequest, InvitationNewRequest, InvitationAcceptRequest, InvitationDeclineRequest, InvitationSasRequest, InvitationDeleteRequest, MessageListRequest, MessageGetRequest, MessageRefreshRequest, MessageDeleteRequest, MessageSendRequest, MessageSendWithAttachmentsRequest, MessageReactRequest, MessageUpdateBodyRequest, MessageSendVoipRequest, StorageListRequest, StorageGetRequest, StorageSetRequest, StorageUnsetRequest, DiscussionStorageListRequest, DiscussionStorageGetRequest, DiscussionStorageSetRequest, DiscussionStorageUnsetRequest]
CommandResponseMessageType = Union[ClientKeyListResponse, ClientKeyGetResponse, ClientKeyNewResponse, ClientKeyDeleteResponse, IdentityListResponse, IdentityAdminGetResponse, IdentityNewResponse, IdentityDeleteResponse, IdentityKeycloakNewResponse, AttachmentListResponse, AttachmentGetResponse, AttachmentDeleteResponse, AttachmentDownloadResponse, ContactListResponse, ContactGetResponse, ContactDeleteResponse, ContactIntroductionResponse, ContactInviteToOneToOneDiscussionResponse, ContactDowngradeOneToOneDiscussionResponse, DiscussionListResponse, DiscussionGetResponse, DiscussionGetByContactResponse, DiscussionGetByGroupResponse, DiscussionEmptyResponse, DiscussionSettingsGetResponse, DiscussionSettingsSetResponse, DiscussionLockedListResponse, DiscussionLockedDeleteResponse, GroupListResponse, GroupGetResponse, GroupNewStandardGroupResponse, GroupNewControlledGroupResponse, GroupNewReadOnlyGroupResponse, GroupNewAdvancedGroupResponse, GroupDisbandResponse, GroupLeaveResponse, GroupUpdateResponse, GroupUnsetPhotoResponse, GroupSetPhotoResponse, IdentityGetResponse, IdentityUpdateDetailsResponse, IdentityRemovePhotoResponse, IdentitySetPhotoResponse, IdentityKeycloakBindResponse, IdentityKeycloakUnbindResponse, IdentitySetApiKeyResponse, IdentitySetConfigurationLinkResponse, InvitationListResponse, InvitationGetResponse, InvitationNewResponse, InvitationAcceptResponse, InvitationDeclineResponse, InvitationSasResponse, InvitationDeleteResponse, MessageListResponse, MessageGetResponse, MessageRefreshResponse, MessageDeleteResponse, MessageSendResponse, MessageSendWithAttachmentsResponse, MessageReactResponse, MessageUpdateBodyResponse, MessageSendVoipResponse, StorageListResponse, StorageGetResponse, StorageSetResponse, StorageUnsetResponse, DiscussionStorageListResponse, DiscussionStorageGetResponse, DiscussionStorageSetResponse, DiscussionStorageUnsetResponse]



import copy
import logging

from ClientHolder import ClientHolder, ClientWrapper
from olvid import datatypes, listeners


async def command_tests(client_1: ClientWrapper, client_2: ClientWrapper):
	discussion_1, discussion_2 = await client_1.get_discussion_associated_to_another_client(client_2), await client_2.get_discussion_associated_to_another_client(client_1)

	await compare_list_and_get_results(client_1)

	# fake bot to avoid watchdogs trigger
	message_count = await count_messages_in_discussion(client_1, discussion_1)
	if message_count > 0:
		message_deleted_bot = client_1.create_notification_bot(listeners.MessageDeletedListener(handler=lambda m: None, count=message_count))

		# empty discussion
		await client_1.discussion_empty(discussion_id=discussion_1.id)

		# wait for messages to be deleted
		await message_deleted_bot.wait_for_listeners_end()

	# check discussion is empty
	assert (await count_messages_in_discussion(client_1, discussion_1)) == 0


async def test_notifications(client_1: ClientWrapper):
	group_name = f"GroupName-{client_1.identity.details.first_name}"
	group_name_updated = f"GroupName-{client_1.identity.details.first_name}-Updated"
	group_description = "NoDescription"

	await compare_list_and_get_results(client_1)

	#####
	# Group creation
	#####
	# create bots to check workflow
	bots = []
	ideal_group = datatypes.Group(name=group_name, description=group_description)
	ideal_discussion = datatypes.Discussion(title=group_name)
	bots.append(client_1.create_notification_bot(listeners.GroupUpdateInProgressListener(handler=lambda gid: gid, count=1)))
	bots.append(client_1.create_notification_bot(listeners.GroupNewListener(handler=client_1.get_check_content_handler(ideal_group, notification_type=listeners.NOTIFICATIONS.GROUP_NEW), count=1)))
	bots.append(client_1.create_notification_bot(listeners.DiscussionNewListener(handler=client_1.get_check_content_handler(ideal_discussion, notification_type=listeners.NOTIFICATIONS.DISCUSSION_NEW), count=1)))
	bots.append(client_1.create_notification_bot(listeners.GroupUpdateFinishedListener(handler=lambda gid: gid, count=1)))

	# create a new group and wait for creation on both sides
	await client_1.group_new_controlled_group(name=group_name, description=group_description, contact_ids=[])

	for bot in bots:
		await bot.wait_for_listeners_end()

	await compare_list_and_get_results(client_1)

	#####
	# update discussion title
	#####
	# TODO use group_new returned group when implemented
	# manually determine new group id
	group_id = 0
	async for g in client_1.group_list():
		group_id = g.id
	assert group_id != 0

	# get the original group
	prev_group: datatypes.Group = await client_1.group_get(group_id=group_id)
	new_group: datatypes.Group = copy.copy(prev_group)
	new_group.name = group_name_updated
	new_group.update_in_progress = False  # set to False to avoid comparison in notification handler

	# create bots to check workflow
	bots = []
	ideal_discussion = datatypes.Discussion(title=new_group.name, group_id=new_group.id)
	ideal_discussion.title = group_name_updated
	bots.append(client_1.create_notification_bot(listeners.GroupUpdateInProgressListener(handler=client_1.get_check_content_handler(new_group.id, notification_type=listeners.NOTIFICATIONS.GROUP_UPDATE_IN_PROGRESS), count=1)))
	bots.append(client_1.create_notification_bot(listeners.GroupNameUpdatedListener(handler=client_1.get_check_content_handler(new_group, group_name, notification_type=listeners.NOTIFICATIONS.GROUP_NEW), count=1)))
	bots.append(client_1.create_notification_bot(listeners.DiscussionTitleUpdatedListener(handler=client_1.get_check_content_handler(ideal_discussion, group_name, notification_type=listeners.NOTIFICATIONS.DISCUSSION_TITLE_UPDATED), count=1)))
	bots.append(client_1.create_notification_bot(listeners.GroupUpdateFinishedListener(handler=client_1.get_check_content_handler(new_group.id, notification_type=listeners.NOTIFICATIONS.GROUP_UPDATE_IN_PROGRESS), count=1)))

	# rename group
	await client_1.group_update(group=new_group)

	for bot in bots:
		await bot.wait_for_listeners_end()

	await compare_list_and_get_results(client_1)

	#####
	# disband group
	#####
	# TODO remove: this is to avoid a known bug: in GroupDeleted notification group.type is not properly set
	new_group.type = datatypes.Group.Type.TYPE_UNSPECIFIED
	# create bots to check workflow
	bots = [
		client_1.create_notification_bot(listeners.GroupUpdateInProgressListener(handler=client_1.get_check_content_handler(new_group.id, notification_type=listeners.NOTIFICATIONS.GROUP_UPDATE_IN_PROGRESS), count=1)),
		client_1.create_notification_bot(listeners.DiscussionLockedListener(handler=client_1.get_check_content_handler(ideal_discussion, notification_type=listeners.NOTIFICATIONS.DISCUSSION_LOCKED), count=1)),
		client_1.create_notification_bot(listeners.GroupDeletedListener(handler=client_1.get_check_content_handler(new_group, notification_type=listeners.NOTIFICATIONS.GROUP_DELETED), count=1)),
		client_1.create_notification_bot(listeners.GroupUpdateFinishedListener(handler=client_1.get_check_content_handler(new_group.id, notification_type=listeners.NOTIFICATIONS.GROUP_UPDATE_IN_PROGRESS), count=1)),
	]

	await client_1.group_disband(group_id=group_id)

	for bot in bots:
		await bot.wait_for_listeners_end()

	await compare_list_and_get_results(client_1)


async def compare_list_and_get_results(client: ClientWrapper):
	async for d_list in client.discussion_list():
		d_get: datatypes.Discussion = await client.discussion_get(discussion_id=d_list.id)
		assert d_get == d_list


async def count_messages_in_discussion(client: ClientWrapper, discussion: datatypes.Discussion):
	message_count = 0
	async for _ in client.message_list(filter=datatypes.MessageFilter(discussion_id=discussion.id)):
		message_count += 1
	return message_count


# noinspection PyUnusedLocal
async def test_discussion(client_holder: ClientHolder, fast_mode: bool = False):
	for c1, c2 in client_holder.get_all_client_pairs():
		logging.info("-" * 10 + " commands " + "-" * 10)
		await command_tests(c1, c2)
		logging.info("-" * 10 + " notifications " + "-" * 10)
		await test_notifications(c1)



from ClientHolder import ClientHolder, ClientWrapper


async def check_entries_count(client: ClientWrapper, expected_count: int, discussion_id: int = None):
	count = 0
	if discussion_id is not None:
		async for _ in client.discussion_storage_list(discussion_id=discussion_id):
			count += 1
	else:
		async for _ in client.storage_list():
			count += 1
	assert count == expected_count


async def check_storage_state(client: ClientWrapper, expected_global_storage: list[tuple[str, str]], discussion_id: int = None, expected_discussion_storage: list[tuple[str, str]] = None):
	for k, v in expected_global_storage:
		value = await client.storage_get(k)
		assert value == v
	await check_entries_count(client, len(expected_global_storage))

	if discussion_id is not None:
		for k, v in expected_discussion_storage:
			value = await client.discussion_storage_get(discussion_id=discussion_id, key=k)
			assert value == v
		await check_entries_count(client, len(expected_discussion_storage), discussion_id=discussion_id)


async def basic_routine_global(client: ClientWrapper):
	# Storage is empty
	await check_storage_state(client, [])

	# set some keys
	test_vectors: list[tuple[str, str]] = [("1", "NoValue"), ("ComplexeK2", "myVaLUASD"), ("123", "42321"), ("EmptyValue", "")]
	for k, v in test_vectors:
		previous_value = await client.storage_set(key=k, value=v)
		assert previous_value == ""

	# override values
	expected_previous_value = test_vectors[0][1]
	test_vectors[0] = (test_vectors[0][0], "NewValue")
	previous_value = await client.storage_set(key=test_vectors[0][0], value=test_vectors[0][1])
	assert previous_value == expected_previous_value
	await check_storage_state(client, test_vectors)

	expected_previous_value = test_vectors[0][1]
	previous_value = await client.storage_set(key=test_vectors[0][0], value=test_vectors[0][1])
	assert previous_value == expected_previous_value
	await check_storage_state(client, test_vectors)

	# check list entry point is ok
	async for element in client.storage_list():
		assert (element.key, element.value) in test_vectors

	# get invalid key
	value = await client.storage_get("UnknownKey")
	assert value == ""
	value = await client.storage_get("ThisKeyIsNotValid")
	assert value == ""

	# delete keys
	for element in test_vectors.copy():
		previous_value = await client.storage_unset(element[0])
		test_vectors.remove(element)
		assert previous_value == element[1]
		await check_storage_state(client, test_vectors)

	await check_storage_state(client, [])


async def basic_routine_discussion(client: ClientWrapper, discussion_id: int):
	# Storage is empty
	await check_storage_state(client, [], discussion_id, [])

	# set some keys
	test_vectors: list[tuple[str, str]] = [("1", "NoValue"), ("ComplexeK2", "myVaLUASD"), ("123", "42321"), ("EmptyValue", "")]
	for k, v in test_vectors:
		previous_value = await client.discussion_storage_set(discussion_id=discussion_id, key=k, value=v)
		assert previous_value == ""

	# override values
	expected_previous_value = test_vectors[0][1]
	test_vectors[0] = (test_vectors[0][0], "NewValue")
	previous_value = await client.discussion_storage_set(discussion_id=discussion_id, key=test_vectors[0][0], value=test_vectors[0][1])
	assert previous_value == expected_previous_value
	await check_storage_state(client, [], discussion_id, test_vectors)

	expected_previous_value = test_vectors[0][1]
	previous_value = await client.discussion_storage_set(discussion_id=discussion_id, key=test_vectors[0][0], value=test_vectors[0][1])
	assert previous_value == expected_previous_value
	await check_storage_state(client, [], discussion_id, test_vectors)

	# check list entry point is ok
	async for element in client.discussion_storage_list(discussion_id=discussion_id):
		assert (element.key, element.value) in test_vectors

	# get invalid key
	value = await client.discussion_storage_get(discussion_id=discussion_id, key="UnknownKey")
	assert value == ""
	value = await client.discussion_storage_get(discussion_id=discussion_id, key="ThisKeyIsNotValid")
	assert value == ""

	# delete keys
	for element in test_vectors.copy():
		previous_value = await client.discussion_storage_unset(discussion_id=discussion_id, key=element[0])
		test_vectors.remove(element)
		assert previous_value == element[1]
		await check_storage_state(client, [], discussion_id, test_vectors)

	await check_storage_state(client, [], discussion_id, [])


async def global_try_to_get_other_client_storage(client_1: ClientWrapper, client_2: ClientWrapper):
	# set elements
	await client_1.storage_set("Client1Key", "YouCantSeeMe")
	await client_2.storage_set("Client2Key", "YouCantSeeMe")

	# get invalid key
	value = await client_1.storage_get("Client2Key")
	assert value == ""

	value = await client_2.storage_get("Client1Key")
	assert value == ""

	# delete keys
	await client_1.storage_unset("Client1Key")
	await client_2.storage_unset("Client2Key")


async def test_storage(client_holder: ClientHolder):
	for c1, c2 in client_holder.get_all_client_pairs():
		await basic_routine_global(c1)
		await basic_routine_global(c2)

		discussion = await c1.get_discussion_associated_to_another_client(other_client=c2)

		await basic_routine_discussion(c1, discussion.id)

		await global_try_to_get_other_client_storage(c1, c2)

		await check_storage_state(c1, [], discussion.id, [])
		await check_storage_state(c2, [])



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: olvid/daemon/notification/v1/message_notifications.proto
# Protobuf Python Version: 5.27.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    3,
    '',
    'olvid/daemon/notification/v1/message_notifications.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from .....olvid.daemon.datatypes.v1 import message_pb2 as olvid_dot_daemon_dot_datatypes_dot_v1_dot_message__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n8olvid/daemon/notification/v1/message_notifications.proto\x12\x1colvid.daemon.notification.v1\x1a\'olvid/daemon/datatypes/v1/message.proto\"(\n&SubscribeToMessageReceivedNotification\"[\n\x1bMessageReceivedNotification\x12<\n\x07message\x18\x01 \x01(\x0b\x32\".olvid.daemon.datatypes.v1.MessageR\x07message\"$\n\"SubscribeToMessageSentNotification\"W\n\x17MessageSentNotification\x12<\n\x07message\x18\x01 \x01(\x0b\x32\".olvid.daemon.datatypes.v1.MessageR\x07message\"\'\n%SubscribeToMessageDeletedNotification\"Z\n\x1aMessageDeletedNotification\x12<\n\x07message\x18\x01 \x01(\x0b\x32\".olvid.daemon.datatypes.v1.MessageR\x07message\"+\n)SubscribeToMessageBodyUpdatedNotification\"\x83\x01\n\x1eMessageBodyUpdatedNotification\x12<\n\x07message\x18\x01 \x01(\x0b\x32\".olvid.daemon.datatypes.v1.MessageR\x07message\x12#\n\rprevious_body\x18\x02 \x01(\tR\x0cpreviousBody\"(\n&SubscribeToMessageUploadedNotification\"[\n\x1bMessageUploadedNotification\x12<\n\x07message\x18\x01 \x01(\x0b\x32\".olvid.daemon.datatypes.v1.MessageR\x07message\")\n\'SubscribeToMessageDeliveredNotification\"\\\n\x1cMessageDeliveredNotification\x12<\n\x07message\x18\x01 \x01(\x0b\x32\".olvid.daemon.datatypes.v1.MessageR\x07message\"$\n\"SubscribeToMessageReadNotification\"W\n\x17MessageReadNotification\x12<\n\x07message\x18\x01 \x01(\x0b\x32\".olvid.daemon.datatypes.v1.MessageR\x07message\"0\n.SubscribeToMessageLocationReceivedNotification\"c\n#MessageLocationReceivedNotification\x12<\n\x07message\x18\x01 \x01(\x0b\x32\".olvid.daemon.datatypes.v1.MessageR\x07message\"4\n2SubscribeToMessageLocationSharingStartNotification\"g\n\'MessageLocationSharingStartNotification\x12<\n\x07message\x18\x01 \x01(\x0b\x32\".olvid.daemon.datatypes.v1.MessageR\x07message\"5\n3SubscribeToMessageLocationSharingUpdateNotification\"\xc1\x01\n(MessageLocationSharingUpdateNotification\x12<\n\x07message\x18\x01 \x01(\x0b\x32\".olvid.daemon.datatypes.v1.MessageR\x07message\x12W\n\x11previous_location\x18\x02 \x01(\x0b\x32*.olvid.daemon.datatypes.v1.MessageLocationR\x10previousLocation\"2\n0SubscribeToMessageLocationSharingEndNotification\"e\n%MessageLocationSharingEndNotification\x12<\n\x07message\x18\x01 \x01(\x0b\x32\".olvid.daemon.datatypes.v1.MessageR\x07message\"-\n+SubscribeToMessageReactionAddedNotification\"\xa8\x01\n MessageReactionAddedNotification\x12<\n\x07message\x18\x01 \x01(\x0b\x32\".olvid.daemon.datatypes.v1.MessageR\x07message\x12\x46\n\x08reaction\x18\x02 \x01(\x0b\x32*.olvid.daemon.datatypes.v1.MessageReactionR\x08reaction\"/\n-SubscribeToMessageReactionUpdatedNotification\"\x83\x02\n\"MessageReactionUpdatedNotification\x12<\n\x07message\x18\x01 \x01(\x0b\x32\".olvid.daemon.datatypes.v1.MessageR\x07message\x12\x46\n\x08reaction\x18\x02 \x01(\x0b\x32*.olvid.daemon.datatypes.v1.MessageReactionR\x08reaction\x12W\n\x11previous_reaction\x18\x03 \x01(\x0b\x32*.olvid.daemon.datatypes.v1.MessageReactionR\x10previousReaction\"/\n-SubscribeToMessageReactionRemovedNotification\"\xaa\x01\n\"MessageReactionRemovedNotification\x12<\n\x07message\x18\x01 \x01(\x0b\x32\".olvid.daemon.datatypes.v1.MessageR\x07message\x12\x46\n\x08reaction\x18\x02 \x01(\x0b\x32*.olvid.daemon.datatypes.v1.MessageReactionR\x08reactionB\xe1\x01\n com.olvid.daemon.notification.v1B\x19MessageNotificationsProtoP\x01Z\x0folvid.io/daemon\xa2\x02\x03ODN\xaa\x02\x1cOlvid.Daemon.Notification.V1\xca\x02\x1cOlvid\\Daemon\\Notification\\V1\xe2\x02(Olvid\\Daemon\\Notification\\V1\\GPBMetadata\xea\x02\x1fOlvid::Daemon::Notification::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'olvid.daemon.notification.v1.message_notifications_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n com.olvid.daemon.notification.v1B\031MessageNotificationsProtoP\001Z\017olvid.io/daemon\242\002\003ODN\252\002\034Olvid.Daemon.Notification.V1\312\002\034Olvid\\Daemon\\Notification\\V1\342\002(Olvid\\Daemon\\Notification\\V1\\GPBMetadata\352\002\037Olvid::Daemon::Notification::V1'
  _globals['_SUBSCRIBETOMESSAGERECEIVEDNOTIFICATION']._serialized_start=131
  _globals['_SUBSCRIBETOMESSAGERECEIVEDNOTIFICATION']._serialized_end=171
  _globals['_MESSAGERECEIVEDNOTIFICATION']._serialized_start=173
  _globals['_MESSAGERECEIVEDNOTIFICATION']._serialized_end=264
  _globals['_SUBSCRIBETOMESSAGESENTNOTIFICATION']._serialized_start=266
  _globals['_SUBSCRIBETOMESSAGESENTNOTIFICATION']._serialized_end=302
  _globals['_MESSAGESENTNOTIFICATION']._serialized_start=304
  _globals['_MESSAGESENTNOTIFICATION']._serialized_end=391
  _globals['_SUBSCRIBETOMESSAGEDELETEDNOTIFICATION']._serialized_start=393
  _globals['_SUBSCRIBETOMESSAGEDELETEDNOTIFICATION']._serialized_end=432
  _globals['_MESSAGEDELETEDNOTIFICATION']._serialized_start=434
  _globals['_MESSAGEDELETEDNOTIFICATION']._serialized_end=524
  _globals['_SUBSCRIBETOMESSAGEBODYUPDATEDNOTIFICATION']._serialized_start=526
  _globals['_SUBSCRIBETOMESSAGEBODYUPDATEDNOTIFICATION']._serialized_end=569
  _globals['_MESSAGEBODYUPDATEDNOTIFICATION']._serialized_start=572
  _globals['_MESSAGEBODYUPDATEDNOTIFICATION']._serialized_end=703
  _globals['_SUBSCRIBETOMESSAGEUPLOADEDNOTIFICATION']._serialized_start=705
  _globals['_SUBSCRIBETOMESSAGEUPLOADEDNOTIFICATION']._serialized_end=745
  _globals['_MESSAGEUPLOADEDNOTIFICATION']._serialized_start=747
  _globals['_MESSAGEUPLOADEDNOTIFICATION']._serialized_end=838
  _globals['_SUBSCRIBETOMESSAGEDELIVEREDNOTIFICATION']._serialized_start=840
  _globals['_SUBSCRIBETOMESSAGEDELIVEREDNOTIFICATION']._serialized_end=881
  _globals['_MESSAGEDELIVEREDNOTIFICATION']._serialized_start=883
  _globals['_MESSAGEDELIVEREDNOTIFICATION']._serialized_end=975
  _globals['_SUBSCRIBETOMESSAGEREADNOTIFICATION']._serialized_start=977
  _globals['_SUBSCRIBETOMESSAGEREADNOTIFICATION']._serialized_end=1013
  _globals['_MESSAGEREADNOTIFICATION']._serialized_start=1015
  _globals['_MESSAGEREADNOTIFICATION']._serialized_end=1102
  _globals['_SUBSCRIBETOMESSAGELOCATIONRECEIVEDNOTIFICATION']._serialized_start=1104
  _globals['_SUBSCRIBETOMESSAGELOCATIONRECEIVEDNOTIFICATION']._serialized_end=1152
  _globals['_MESSAGELOCATIONRECEIVEDNOTIFICATION']._serialized_start=1154
  _globals['_MESSAGELOCATIONRECEIVEDNOTIFICATION']._serialized_end=1253
  _globals['_SUBSCRIBETOMESSAGELOCATIONSHARINGSTARTNOTIFICATION']._serialized_start=1255
  _globals['_SUBSCRIBETOMESSAGELOCATIONSHARINGSTARTNOTIFICATION']._serialized_end=1307
  _globals['_MESSAGELOCATIONSHARINGSTARTNOTIFICATION']._serialized_start=1309
  _globals['_MESSAGELOCATIONSHARINGSTARTNOTIFICATION']._serialized_end=1412
  _globals['_SUBSCRIBETOMESSAGELOCATIONSHARINGUPDATENOTIFICATION']._serialized_start=1414
  _globals['_SUBSCRIBETOMESSAGELOCATIONSHARINGUPDATENOTIFICATION']._serialized_end=1467
  _globals['_MESSAGELOCATIONSHARINGUPDATENOTIFICATION']._serialized_start=1470
  _globals['_MESSAGELOCATIONSHARINGUPDATENOTIFICATION']._serialized_end=1663
  _globals['_SUBSCRIBETOMESSAGELOCATIONSHARINGENDNOTIFICATION']._serialized_start=1665
  _globals['_SUBSCRIBETOMESSAGELOCATIONSHARINGENDNOTIFICATION']._serialized_end=1715
  _globals['_MESSAGELOCATIONSHARINGENDNOTIFICATION']._serialized_start=1717
  _globals['_MESSAGELOCATIONSHARINGENDNOTIFICATION']._serialized_end=1818
  _globals['_SUBSCRIBETOMESSAGEREACTIONADDEDNOTIFICATION']._serialized_start=1820
  _globals['_SUBSCRIBETOMESSAGEREACTIONADDEDNOTIFICATION']._serialized_end=1865
  _globals['_MESSAGEREACTIONADDEDNOTIFICATION']._serialized_start=1868
  _globals['_MESSAGEREACTIONADDEDNOTIFICATION']._serialized_end=2036
  _globals['_SUBSCRIBETOMESSAGEREACTIONUPDATEDNOTIFICATION']._serialized_start=2038
  _globals['_SUBSCRIBETOMESSAGEREACTIONUPDATEDNOTIFICATION']._serialized_end=2085
  _globals['_MESSAGEREACTIONUPDATEDNOTIFICATION']._serialized_start=2088
  _globals['_MESSAGEREACTIONUPDATEDNOTIFICATION']._serialized_end=2347
  _globals['_SUBSCRIBETOMESSAGEREACTIONREMOVEDNOTIFICATION']._serialized_start=2349
  _globals['_SUBSCRIBETOMESSAGEREACTIONREMOVEDNOTIFICATION']._serialized_end=2396
  _globals['_MESSAGEREACTIONREMOVEDNOTIFICATION']._serialized_start=2399
  _globals['_MESSAGEREACTIONREMOVEDNOTIFICATION']._serialized_end=2569
# @@protoc_insertion_point(module_scope)



from .commands import AttachmentListRequest, AttachmentListResponse, AttachmentGetRequest, AttachmentGetResponse, AttachmentDeleteRequest, AttachmentDeleteResponse, AttachmentDownloadRequest, AttachmentDownloadResponse, ContactListRequest, ContactListResponse, ContactGetRequest, ContactGetResponse, ContactDeleteRequest, ContactDeleteResponse, ContactIntroductionRequest, ContactIntroductionResponse, ContactInviteToOneToOneDiscussionRequest, ContactInviteToOneToOneDiscussionResponse, ContactDowngradeOneToOneDiscussionRequest, ContactDowngradeOneToOneDiscussionResponse, DiscussionListRequest, DiscussionListResponse, DiscussionGetRequest, DiscussionGetResponse, DiscussionGetByContactRequest, DiscussionGetByContactResponse, DiscussionGetByGroupRequest, DiscussionGetByGroupResponse, DiscussionEmptyRequest, DiscussionEmptyResponse, DiscussionSettingsGetRequest, DiscussionSettingsGetResponse, DiscussionSettingsSetRequest, DiscussionSettingsSetResponse, DiscussionLockedListRequest, DiscussionLockedListResponse, DiscussionLockedDeleteRequest, DiscussionLockedDeleteResponse, GroupListRequest, GroupListResponse, GroupGetRequest, GroupGetResponse, GroupNewStandardGroupRequest, GroupNewStandardGroupResponse, GroupNewControlledGroupRequest, GroupNewControlledGroupResponse, GroupNewReadOnlyGroupRequest, GroupNewReadOnlyGroupResponse, GroupNewAdvancedGroupRequest, GroupNewAdvancedGroupResponse, GroupDisbandRequest, GroupDisbandResponse, GroupLeaveRequest, GroupLeaveResponse, GroupUpdateRequest, GroupUpdateResponse, GroupUnsetPhotoRequest, GroupUnsetPhotoResponse, GroupSetPhotoRequestMetadata, GroupSetPhotoRequest, GroupSetPhotoResponse, IdentityGetRequest, IdentityGetResponse, IdentityUpdateDetailsRequest, IdentityUpdateDetailsResponse, IdentityRemovePhotoRequest, IdentityRemovePhotoResponse, IdentitySetPhotoRequestMetadata, IdentitySetPhotoRequest, IdentitySetPhotoResponse, IdentityKeycloakBindRequest, IdentityKeycloakBindResponse, IdentityKeycloakUnbindRequest, IdentityKeycloakUnbindResponse, IdentitySetApiKeyRequest, IdentitySetApiKeyResponse, IdentitySetConfigurationLinkRequest, IdentitySetConfigurationLinkResponse, InvitationListRequest, InvitationListResponse, InvitationGetRequest, InvitationGetResponse, InvitationNewRequest, InvitationNewResponse, InvitationAcceptRequest, InvitationAcceptResponse, InvitationDeclineRequest, InvitationDeclineResponse, InvitationSasRequest, InvitationSasResponse, InvitationDeleteRequest, InvitationDeleteResponse, MessageListRequest, MessageListResponse, MessageGetRequest, MessageGetResponse, MessageRefreshRequest, MessageRefreshResponse, MessageDeleteRequest, MessageDeleteResponse, MessageSendRequest, MessageSendResponse, MessageSendWithAttachmentsRequest, MessageSendWithAttachmentsRequestMetadata, MessageSendWithAttachmentsResponse, MessageReactRequest, MessageReactResponse, MessageUpdateBodyRequest, MessageUpdateBodyResponse, MessageSendVoipRequest, MessageSendVoipResponse, StorageListRequest, StorageListResponse, StorageGetRequest, StorageGetResponse, StorageSetRequest, StorageSetResponse, StorageUnsetRequest, StorageUnsetResponse, DiscussionStorageListRequest, DiscussionStorageListResponse, DiscussionStorageGetRequest, DiscussionStorageGetResponse, DiscussionStorageSetRequest, DiscussionStorageSetResponse, DiscussionStorageUnsetRequest, DiscussionStorageUnsetResponse, IdentityCommandServiceStub, InvitationCommandServiceStub, ContactCommandServiceStub, GroupCommandServiceStub, DiscussionCommandServiceStub, MessageCommandServiceStub, AttachmentCommandServiceStub, StorageCommandServiceStub, DiscussionStorageCommandServiceStub
del commands



from asyncio import iscoroutine
from typing import Callable, Coroutine, Union, Any

from grpc.aio import AioRpcError

from .logger import tools_logger
from ..core.OlvidClient import OlvidClient
from ..core.OlvidBot import OlvidBot
from .. import datatypes
from ..listeners.ListenersImplementation import MessageReceivedListener, MessageSentListener, AttachmentUploadedListener, MessageUploadedListener


class SelfCleaningBot(OlvidBot):
	# with no parameters we clean in and out messages
	# if you specify is_message_for_cleaning parameter we disable in and out clean,
	# we only use is_message_for_cleaning method (so be sure to cover every case).
	def __init__(self, parent_client: OlvidClient = None, clean_inbound_messages: bool = None, clean_outbound_messages: bool = None, is_message_for_cleaning: Callable[[datatypes.Message], Union[bool, Coroutine[Any, Any, bool]]] = None):
		super().__init__(parent_client=parent_client)

		# keep original messages (sent by message_sent notification)
		# then we decrement attachment_number field every time an attachment had been uploaded
		# when every attachment had been sent we delete message and attachment
		self._pending_outbound_messages_by_id: dict[int, datatypes.Message] = {}

		self.is_message_for_cleaning: Callable[[datatypes.Message], Union[Coroutine[Any, Any, bool]]] = self._is_message_for_cleaning_wrapper(is_message_for_cleaning) if is_message_for_cleaning else None
		if self.is_message_for_cleaning and (clean_inbound_messages is not None or clean_outbound_messages is not None):
			raise ValueError("Cannot set is_message_for_cleaning and clean_inbound_messages or clean_outbound_messages")
		self.clean_inbound_messages: bool = clean_inbound_messages if clean_inbound_messages is not None else True
		self.clean_outbound_messages: bool = clean_outbound_messages if clean_outbound_messages is not None else True

		self.add_listener(listener=MessageReceivedListener(handler=self.message_received_handler))
		self.add_listener(listener=MessageSentListener(handler=self.message_sent_handler))
		self.add_listener(listener=MessageUploadedListener(handler=self.message_uploaded_handler))
		self.add_listener(listener=AttachmentUploadedListener(handler=self.attachment_uploaded_handler))

		# clean on every start
		self.add_background_task(self.clean_on_start_task(client=self, clean_inbound_messages=self.clean_inbound_messages, clean_outbound_messages=self.clean_outbound_messages, is_message_for_cleaning=self.is_message_for_cleaning))

	# this is static to use it even if we do not need a full SelfCleaningBot
	@staticmethod
	async def clean_on_start_task(client: OlvidClient, clean_inbound_messages: bool = None, clean_outbound_messages: bool = None, is_message_for_cleaning: Callable[[datatypes.Message], Union[Coroutine[Any, Any, bool]]] = None):
		try:
			# delete every messages
			deleted_messages_count = 0
			async for message in client.message_list():
				if is_message_for_cleaning and await is_message_for_cleaning(message):
					await client.message_delete(message_id=message.id)
					deleted_messages_count += 1
				elif clean_inbound_messages and message.is_inbound():
					await client.message_delete(message_id=message.id)
					deleted_messages_count += 1
				elif clean_outbound_messages and message.is_outbound():
					await client.message_delete(message_id=message.id)
					deleted_messages_count += 1

			# delete locked discussions
			locked_discussions_count = 0
			async for locked_discussion in client.discussion_locked_list():
				locked_discussions_count += 1
				await client.discussion_locked_delete(discussion_id=locked_discussion.id)

			tools_logger.debug(f"{client.__class__.__name__}: start clean: deleted {deleted_messages_count} messages and {locked_discussions_count} locked discussion")
		except AioRpcError as rpc_error:
			tools_logger.error(f"{client.__class__.__name__}: clean on start: {rpc_error.code()}: {rpc_error.details()}")
		except Exception:
			tools_logger.exception(f"{client.__class__.__name__}: clean on start: unexpected error")

	# delete every received messages
	async def message_received_handler(self, message: datatypes.Message):
		if self.clean_inbound_messages or (self.is_message_for_cleaning and await self.is_message_for_cleaning(message)):
			try:
				await self.message_delete(message_id=message.id)
			except Exception:
				tools_logger.exception(f"{self.name}: on_message_received: unexpected exception")

	# delete outbound messages when they arrive or when all attachments have been uploaded
	async def message_sent_handler(self, message: datatypes.Message):
		if self.clean_outbound_messages or (self.is_message_for_cleaning and await self.is_message_for_cleaning(message)):
			self._pending_outbound_messages_by_id[message.id.id] = message

	async def message_uploaded_handler(self, message: datatypes.Message):
		try:
			message = self._pending_outbound_messages_by_id.get(message.id.id)
			if not message:
				return
			# wait for attachments to be uploaded before deletion
			if message.attachments_count > 0:
				return

			await self.message_delete(message_id=message.id)
			self._pending_outbound_messages_by_id.pop(message.id.id)
		except Exception:
			tools_logger.exception(f"{self.name}: on_message_uploaded: unexpected exception")

	async def attachment_uploaded_handler(self, attachment: datatypes.Attachment):
		try:
			message = self._pending_outbound_messages_by_id.get(attachment.message_id.id)
			if not message:
				return
			message.attachments_count -= 1

			# every attachment were uploaded
			if message.attachments_count == 0:
				await self.message_delete(message_id=message.id)
				self._pending_outbound_messages_by_id.pop(message.id.id)
		except Exception:
			tools_logger.exception(f"{self.name}: on_attachment_uploaded: unexpected exception")

	@staticmethod
	def _is_message_for_cleaning_wrapper(is_message_for_cleaning: Callable[[datatypes.Message], Union[bool, Coroutine[Any, Any, bool]]]) -> Callable[[datatypes.Message], Union[Coroutine[Any, Any, bool]]]:
		async def wrapper(message: datatypes.Message):
			res = is_message_for_cleaning(message)
			if iscoroutine(res):
				res = await res
			return res
		return wrapper

	####
	# tools
	####
	# use this just after sending a message to manually add it to
	def delete_message_when_uploaded(self, message: datatypes.Message):
		self._pending_outbound_messages_by_id[message.id.id] = message



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: olvid/daemon/services/v1/notification_service.proto
# Protobuf Python Version: 5.27.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    3,
    '',
    'olvid/daemon/services/v1/notification_service.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from .....olvid.daemon.notification.v1 import attachment_notifications_pb2 as olvid_dot_daemon_dot_notification_dot_v1_dot_attachment__notifications__pb2
from .....olvid.daemon.notification.v1 import contact_notifications_pb2 as olvid_dot_daemon_dot_notification_dot_v1_dot_contact__notifications__pb2
from .....olvid.daemon.notification.v1 import discussion_notifications_pb2 as olvid_dot_daemon_dot_notification_dot_v1_dot_discussion__notifications__pb2
from .....olvid.daemon.notification.v1 import group_notifications_pb2 as olvid_dot_daemon_dot_notification_dot_v1_dot_group__notifications__pb2
from .....olvid.daemon.notification.v1 import invitation_notifications_pb2 as olvid_dot_daemon_dot_notification_dot_v1_dot_invitation__notifications__pb2
from .....olvid.daemon.notification.v1 import message_notifications_pb2 as olvid_dot_daemon_dot_notification_dot_v1_dot_message__notifications__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n3olvid/daemon/services/v1/notification_service.proto\x12\x18olvid.daemon.services.v1\x1a;olvid/daemon/notification/v1/attachment_notifications.proto\x1a\x38olvid/daemon/notification/v1/contact_notifications.proto\x1a;olvid/daemon/notification/v1/discussion_notifications.proto\x1a\x36olvid/daemon/notification/v1/group_notifications.proto\x1a;olvid/daemon/notification/v1/invitation_notifications.proto\x1a\x38olvid/daemon/notification/v1/message_notifications.proto2\x8d\x05\n\x1dInvitationNotificationService\x12\x9d\x01\n\x12InvitationReceived\x12G.olvid.daemon.notification.v1.SubscribeToInvitationReceivedNotification\x1a<.olvid.daemon.notification.v1.InvitationReceivedNotification0\x01\x12\x91\x01\n\x0eInvitationSent\x12\x43.olvid.daemon.notification.v1.SubscribeToInvitationSentNotification\x1a\x38.olvid.daemon.notification.v1.InvitationSentNotification0\x01\x12\x9a\x01\n\x11InvitationDeleted\x12\x46.olvid.daemon.notification.v1.SubscribeToInvitationDeletedNotification\x1a;.olvid.daemon.notification.v1.InvitationDeletedNotification0\x01\x12\x9a\x01\n\x11InvitationUpdated\x12\x46.olvid.daemon.notification.v1.SubscribeToInvitationUpdatedNotification\x1a;.olvid.daemon.notification.v1.InvitationUpdatedNotification0\x01\x32\xe7\x03\n\x1a\x43ontactNotificationService\x12\x87\x01\n\nContactNew\x12?.olvid.daemon.notification.v1.SubscribeToContactNewNotification\x1a\x34.olvid.daemon.notification.v1.ContactNewNotification\"\x00\x30\x01\x12\x93\x01\n\x0e\x43ontactDeleted\x12\x43.olvid.daemon.notification.v1.SubscribeToContactDeletedNotification\x1a\x38.olvid.daemon.notification.v1.ContactDeletedNotification\"\x00\x30\x01\x12\xa8\x01\n\x15\x43ontactDetailsUpdated\x12J.olvid.daemon.notification.v1.SubscribeToContactDetailsUpdatedNotification\x1a?.olvid.daemon.notification.v1.ContactDetailsUpdatedNotification\"\x00\x30\x01\x32\xe8\x0f\n\x18GroupNotificationService\x12\x81\x01\n\x08GroupNew\x12=.olvid.daemon.notification.v1.SubscribeToGroupNewNotification\x1a\x32.olvid.daemon.notification.v1.GroupNewNotification\"\x00\x30\x01\x12\x8d\x01\n\x0cGroupDeleted\x12\x41.olvid.daemon.notification.v1.SubscribeToGroupDeletedNotification\x1a\x36.olvid.daemon.notification.v1.GroupDeletedNotification\"\x00\x30\x01\x12\x99\x01\n\x10GroupNameUpdated\x12\x45.olvid.daemon.notification.v1.SubscribeToGroupNameUpdatedNotification\x1a:.olvid.daemon.notification.v1.GroupNameUpdatedNotification\"\x00\x30\x01\x12\xae\x01\n\x17GroupDescriptionUpdated\x12L.olvid.daemon.notification.v1.SubscribeToGroupDescriptionUpdatedNotification\x1a\x41.olvid.daemon.notification.v1.GroupDescriptionUpdatedNotification\"\x00\x30\x01\x12\xae\x01\n\x17GroupPendingMemberAdded\x12L.olvid.daemon.notification.v1.SubscribeToGroupPendingMemberAddedNotification\x1a\x41.olvid.daemon.notification.v1.GroupPendingMemberAddedNotification\"\x00\x30\x01\x12\xb4\x01\n\x19GroupPendingMemberRemoved\x12N.olvid.daemon.notification.v1.SubscribeToGroupPendingMemberRemovedNotification\x1a\x43.olvid.daemon.notification.v1.GroupPendingMemberRemovedNotification\"\x00\x30\x01\x12\x9c\x01\n\x11GroupMemberJoined\x12\x46.olvid.daemon.notification.v1.SubscribeToGroupMemberJoinedNotification\x1a;.olvid.daemon.notification.v1.GroupMemberJoinedNotification\"\x00\x30\x01\x12\x96\x01\n\x0fGroupMemberLeft\x12\x44.olvid.daemon.notification.v1.SubscribeToGroupMemberLeftNotification\x1a\x39.olvid.daemon.notification.v1.GroupMemberLeftNotification\"\x00\x30\x01\x12\xb7\x01\n\x1aGroupOwnPermissionsUpdated\x12O.olvid.daemon.notification.v1.SubscribeToGroupOwnPermissionsUpdatedNotification\x1a\x44.olvid.daemon.notification.v1.GroupOwnPermissionsUpdatedNotification\"\x00\x30\x01\x12\xc0\x01\n\x1dGroupMemberPermissionsUpdated\x12R.olvid.daemon.notification.v1.SubscribeToGroupMemberPermissionsUpdatedNotification\x1aG.olvid.daemon.notification.v1.GroupMemberPermissionsUpdatedNotification\"\x00\x30\x01\x12\xa8\x01\n\x15GroupUpdateInProgress\x12J.olvid.daemon.notification.v1.SubscribeToGroupUpdateInProgressNotification\x1a?.olvid.daemon.notification.v1.GroupUpdateInProgressNotification\"\x00\x30\x01\x12\xa2\x01\n\x13GroupUpdateFinished\x12H.olvid.daemon.notification.v1.SubscribeToGroupUpdateFinishedNotification\x1a=.olvid.daemon.notification.v1.GroupUpdateFinishedNotification\"\x00\x30\x01\x32\xb3\x05\n\x1d\x44iscussionNotificationService\x12\x90\x01\n\rDiscussionNew\x12\x42.olvid.daemon.notification.v1.SubscribeToDiscussionNewNotification\x1a\x37.olvid.daemon.notification.v1.DiscussionNewNotification\"\x00\x30\x01\x12\x99\x01\n\x10\x44iscussionLocked\x12\x45.olvid.daemon.notification.v1.SubscribeToDiscussionLockedNotification\x1a:.olvid.daemon.notification.v1.DiscussionLockedNotification\"\x00\x30\x01\x12\xab\x01\n\x16\x44iscussionTitleUpdated\x12K.olvid.daemon.notification.v1.SubscribeToDiscussionTitleUpdatedNotification\x1a@.olvid.daemon.notification.v1.DiscussionTitleUpdatedNotification\"\x00\x30\x01\x12\xb4\x01\n\x19\x44iscussionSettingsUpdated\x12N.olvid.daemon.notification.v1.SubscribeToDiscussionSettingsUpdatedNotification\x1a\x43.olvid.daemon.notification.v1.DiscussionSettingsUpdatedNotification\"\x00\x30\x01\x32\xa5\x12\n\x1aMessageNotificationService\x12\x96\x01\n\x0fMessageReceived\x12\x44.olvid.daemon.notification.v1.SubscribeToMessageReceivedNotification\x1a\x39.olvid.daemon.notification.v1.MessageReceivedNotification\"\x00\x30\x01\x12\x8a\x01\n\x0bMessageSent\x12@.olvid.daemon.notification.v1.SubscribeToMessageSentNotification\x1a\x35.olvid.daemon.notification.v1.MessageSentNotification\"\x00\x30\x01\x12\x93\x01\n\x0eMessageDeleted\x12\x43.olvid.daemon.notification.v1.SubscribeToMessageDeletedNotification\x1a\x38.olvid.daemon.notification.v1.MessageDeletedNotification\"\x00\x30\x01\x12\x9f\x01\n\x12MessageBodyUpdated\x12G.olvid.daemon.notification.v1.SubscribeToMessageBodyUpdatedNotification\x1a<.olvid.daemon.notification.v1.MessageBodyUpdatedNotification\"\x00\x30\x01\x12\x96\x01\n\x0fMessageUploaded\x12\x44.olvid.daemon.notification.v1.SubscribeToMessageUploadedNotification\x1a\x39.olvid.daemon.notification.v1.MessageUploadedNotification\"\x00\x30\x01\x12\x99\x01\n\x10MessageDelivered\x12\x45.olvid.daemon.notification.v1.SubscribeToMessageDeliveredNotification\x1a:.olvid.daemon.notification.v1.MessageDeliveredNotification\"\x00\x30\x01\x12\x8a\x01\n\x0bMessageRead\x12@.olvid.daemon.notification.v1.SubscribeToMessageReadNotification\x1a\x35.olvid.daemon.notification.v1.MessageReadNotification\"\x00\x30\x01\x12\xae\x01\n\x17MessageLocationReceived\x12L.olvid.daemon.notification.v1.SubscribeToMessageLocationReceivedNotification\x1a\x41.olvid.daemon.notification.v1.MessageLocationReceivedNotification\"\x00\x30\x01\x12\xba\x01\n\x1bMessageLocationSharingStart\x12P.olvid.daemon.notification.v1.SubscribeToMessageLocationSharingStartNotification\x1a\x45.olvid.daemon.notification.v1.MessageLocationSharingStartNotification\"\x00\x30\x01\x12\xbd\x01\n\x1cMessageLocationSharingUpdate\x12Q.olvid.daemon.notification.v1.SubscribeToMessageLocationSharingUpdateNotification\x1a\x46.olvid.daemon.notification.v1.MessageLocationSharingUpdateNotification\"\x00\x30\x01\x12\xb4\x01\n\x19MessageLocationSharingEnd\x12N.olvid.daemon.notification.v1.SubscribeToMessageLocationSharingEndNotification\x1a\x43.olvid.daemon.notification.v1.MessageLocationSharingEndNotification\"\x00\x30\x01\x12\xa5\x01\n\x14MessageReactionAdded\x12I.olvid.daemon.notification.v1.SubscribeToMessageReactionAddedNotification\x1a>.olvid.daemon.notification.v1.MessageReactionAddedNotification\"\x00\x30\x01\x12\xab\x01\n\x16MessageReactionUpdated\x12K.olvid.daemon.notification.v1.SubscribeToMessageReactionUpdatedNotification\x1a@.olvid.daemon.notification.v1.MessageReactionUpdatedNotification\"\x00\x30\x01\x12\xab\x01\n\x16MessageReactionRemoved\x12K.olvid.daemon.notification.v1.SubscribeToMessageReactionRemovedNotification\x1a@.olvid.daemon.notification.v1.MessageReactionRemovedNotification\"\x00\x30\x01\x32\xe3\x02\n\x1d\x41ttachmentNotificationService\x12\x9f\x01\n\x12\x41ttachmentReceived\x12G.olvid.daemon.notification.v1.SubscribeToAttachmentReceivedNotification\x1a<.olvid.daemon.notification.v1.AttachmentReceivedNotification\"\x00\x30\x01\x12\x9f\x01\n\x12\x41ttachmentUploaded\x12G.olvid.daemon.notification.v1.SubscribeToAttachmentUploadedNotification\x1a<.olvid.daemon.notification.v1.AttachmentUploadedNotification\"\x00\x30\x01\x42\xcc\x01\n\x1c\x63om.olvid.daemon.services.v1B\x18NotificationServiceProtoP\x01Z\x0folvid.io/daemon\xa2\x02\x03ODS\xaa\x02\x18Olvid.Daemon.Services.V1\xca\x02\x18Olvid\\Daemon\\Services\\V1\xe2\x02$Olvid\\Daemon\\Services\\V1\\GPBMetadata\xea\x02\x1bOlvid::Daemon::Services::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'olvid.daemon.services.v1.notification_service_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\034com.olvid.daemon.services.v1B\030NotificationServiceProtoP\001Z\017olvid.io/daemon\242\002\003ODS\252\002\030Olvid.Daemon.Services.V1\312\002\030Olvid\\Daemon\\Services\\V1\342\002$Olvid\\Daemon\\Services\\V1\\GPBMetadata\352\002\033Olvid::Daemon::Services::V1'
  _globals['_INVITATIONNOTIFICATIONSERVICE']._serialized_start=437
  _globals['_INVITATIONNOTIFICATIONSERVICE']._serialized_end=1090
  _globals['_CONTACTNOTIFICATIONSERVICE']._serialized_start=1093
  _globals['_CONTACTNOTIFICATIONSERVICE']._serialized_end=1580
  _globals['_GROUPNOTIFICATIONSERVICE']._serialized_start=1583
  _globals['_GROUPNOTIFICATIONSERVICE']._serialized_end=3607
  _globals['_DISCUSSIONNOTIFICATIONSERVICE']._serialized_start=3610
  _globals['_DISCUSSIONNOTIFICATIONSERVICE']._serialized_end=4301
  _globals['_MESSAGENOTIFICATIONSERVICE']._serialized_start=4304
  _globals['_MESSAGENOTIFICATIONSERVICE']._serialized_end=6645
  _globals['_ATTACHMENTNOTIFICATIONSERVICE']._serialized_start=6648
  _globals['_ATTACHMENTNOTIFICATIONSERVICE']._serialized_end=7003
# @@protoc_insertion_point(module_scope)



# import core clients
# noinspection PyUnresolvedReferences
from .core import OlvidClient
# noinspection PyUnresolvedReferences
from .core import OlvidBot
# noinspection PyUnresolvedReferences
from .core import OlvidAdminClient

# import core elements
# noinspection PyUnresolvedReferences
from . import listeners

# import overlay modules
# noinspection PyUnresolvedReferences
from . import datatypes
# noinspection PyUnresolvedReferences
from . import internal

# import bots
# noinspection PyUnresolvedReferences
from . import tools

# delete imported modules
if "core" in locals() or "core" in globals():
	del core

if "protobuf" in locals() or "protobuf" in globals():
	# noinspection PyUnresolvedReferences
	del protobuf

# noinspection PyUnresolvedReferences
from .version import __version__
del version



from grpc.aio import AioRpcError

from .logger import tools_logger
from ..core.OlvidClient import OlvidClient
from ..core.OlvidBot import OlvidBot
from .. import datatypes


class AutoInvitationBot(OlvidBot):
	def __init__(self, parent_client: OlvidClient = None):
		super().__init__(parent_client=parent_client)

		# accept pending invitations (need a background task)
		async def accept_invitations_task():
			try:
				async for invitation in self.invitation_list():
					await self.accept_invitation(invitation)
			except AioRpcError as rpc_error:
				tools_logger.error(f"{self.name}: accept invitation on start task: {rpc_error.code()}: {rpc_error.details()}")
			except Exception:
				tools_logger.exception(f"{self.name}: accept invitation on start task: unexpected error")

		self.add_background_task(coroutine=accept_invitations_task(), name=f"{self.__class__.__name__}-accept_invitations_task")

	async def on_invitation_received(self, invitation: datatypes.Invitation):
		await self.accept_invitation(invitation=invitation)

	async def accept_invitation(self, invitation: datatypes.Invitation):
		if invitation.status in [
			datatypes.Invitation.Status.STATUS_INVITATION_WAIT_YOU_TO_ACCEPT,
			datatypes.Invitation.Status.STATUS_INTRODUCTION_WAIT_YOU_TO_ACCEPT,
			datatypes.Invitation.Status.STATUS_ONE_TO_ONE_INVITATION_WAIT_YOU_TO_ACCEPT,
			datatypes.Invitation.Status.STATUS_GROUP_INVITATION_WAIT_YOU_TO_ACCEPT]:
			tools_logger.info(f"{self.__class__.__name__}: invitation accepted: {invitation.id}: {invitation.status}")
			await self.invitation_accept(invitation_id=invitation.id)



import asyncio
import os
import json
from asyncio import iscoroutine

from typing import Callable, Optional, Coroutine, Union, Any

# for compatibility with python 3.10
from typing import TypeVar
# noinspection PyTypeHints
Self = TypeVar("MonitoringBot")

from .logger import tools_logger

# if aiohttp is not installed we use urllib run in executor to avoid adding dependencies for olvid module
try:
	import aiohttp
	tools_logger.debug("MonitoringBot: using aiohttp")
except ImportError:
	import urllib.request
	import urllib.error
	tools_logger.debug("MonitoringBot: using urllib")

from ..core.OlvidClient import OlvidClient
from ..core.OlvidBot import OlvidBot


class MonitoringBot(OlvidBot):
	# mandatory
	MONITORING_ENABLE_ENV_NAME = "MONITORING_ENABLE"
	MONITORING_SERVER_ENV_NAME = "MONITORING_SERVER"
	MONITORING_BOT_NAME_ENV_NAME = "MONITORING_BOT_NAME"
	# optional (at least one option must be set)
	MONITORING_MAIL_ADDRESSES_ENV_NAME = "MONITORING_MAIL_ADDRESSES"
	MONITORING_SMS_NUMBERS_ENV_NAME = "MONITORING_SMS_NUMBERS"
	MONITORING_WEBHOOK_URLS_ENV_NAME = "MONITORING_WEBHOOK_URLS"
	# optional
	# send a test notification on bot test (default to False)
	MONITORING_SEND_TEST_NOTIFICATION_ENV_NAME = "MONITORING_SEND_TEST_NOTIFICATION"
	# time in seconds between two test routine (default to 300 seconds)
	MONITORING_PERIOD_ENV_NAME = "MONITORING_PERIOD"

	def __init__(self, parent_client: OlvidClient = None, monitoring_routine: Callable[[Self], Union[bool, Coroutine[Any, Any, bool]]] = None):
		super().__init__(parent_client=parent_client)

		self.argument_monitoring_routine: Callable[[Self], Union[bool, Coroutine[Any, Any, bool]]] = monitoring_routine

		# init aiohttp session if using it
		self.aiohttp_session: Optional[aiohttp.ClientSession] = None
		if "aiohttp" in globals():
			self.aiohttp_session = aiohttp.ClientSession()

		# return None if monitoring is disabled
		self._monitoring_configuration: Optional[MonitoringBot.MonitoringConfiguration] = self._parse_env_config()
		if not self._monitoring_configuration:
			return

		if len(self._monitoring_configuration.sms_numbers) == 0 and len(self._monitoring_configuration.webhook_urls) == 0:
			tools_logger.warning(f"{self.name}: We do not recommend using mail alerts on it's own cause you have to check your mail box to validate your subscription on every start")

		subscribers_count = len(self._monitoring_configuration.sms_numbers) if self._monitoring_configuration.sms_numbers else 0
		subscribers_count += len(self._monitoring_configuration.mail_addresses) if self._monitoring_configuration.mail_addresses else 0
		subscribers_count += len(self._monitoring_configuration.webhook_urls) if self._monitoring_configuration.webhook_urls else 0
		tools_logger.info(f"{self.name}: enabled monitoring: {subscribers_count} alerts recipient set")

		# start monitoring
		self.add_background_task(self._start_monitoring_process(), f"{self.name}-initial-start-monitoring-task")

	async def stop(self):
		await super().stop()
		if self.aiohttp_session is not None:
			await self.aiohttp_session.close()

	#####
	# Interface (method to override)
	#####
	# override this method to execute code before calling refresh entry point
	# if this method fails, refresh will not be call and bot will be considered as down
	async def monitoring_routine(self):
		pass

	#####
	# Configuration parsing
	#####
	def _parse_env_config(self) -> Optional["MonitoringConfiguration"]:
		# check monitoring is enabled
		if not os.getenv(self.MONITORING_ENABLE_ENV_NAME, False):
			return None

		# check required config (server and bot name)
		server_url = os.getenv(self.MONITORING_SERVER_ENV_NAME, None)
		bot_name = os.getenv(self.MONITORING_BOT_NAME_ENV_NAME, None)
		if not server_url or not bot_name:
			raise self.MonitoringConfiguration.InvalidMonitoringConfigurationException(f"Monitoring server or bot name not specified")

		mail_addresses: list[str] = []
		if os.getenv(self.MONITORING_MAIL_ADDRESSES_ENV_NAME, ""):
			for mail_address in os.getenv(self.MONITORING_MAIL_ADDRESSES_ENV_NAME, "").split(","):
				if not mail_address:
					raise self.MonitoringConfiguration.InvalidMonitoringConfigurationException(f"Invalid mail address: {mail_address}")
				mail_addresses.append(mail_address)

		sms_numbers: list[str] = []
		if os.getenv(self.MONITORING_SMS_NUMBERS_ENV_NAME, ""):
			for sms_number in os.getenv(self.MONITORING_SMS_NUMBERS_ENV_NAME, "").split(","):
				if not sms_number:
					raise self.MonitoringConfiguration.InvalidMonitoringConfigurationException(f"Invalid mail address: {sms_number}")
				sms_numbers.append(sms_number)

		webhook_urls: list[str] = []
		if os.getenv(self.MONITORING_WEBHOOK_URLS_ENV_NAME, ""):
			for webhook_url in os.getenv(self.MONITORING_WEBHOOK_URLS_ENV_NAME, "").split(","):
				if not webhook_url:
					raise self.MonitoringConfiguration.InvalidMonitoringConfigurationException(f"Invalid webhook url: {webhook_url}")
				webhook_urls.append(webhook_url)

		if not mail_addresses and not sms_numbers and not webhook_urls:
			raise self.MonitoringConfiguration.InvalidMonitoringConfigurationException(f"No monitoring alert recipient specified")

		# get optional elements
		send_test_notification = os.getenv(self.MONITORING_SEND_TEST_NOTIFICATION_ENV_NAME, False)
		try:
			period = int(os.getenv(self.MONITORING_PERIOD_ENV_NAME, "300"))
		except ValueError:
			raise self.MonitoringConfiguration.InvalidMonitoringConfigurationException(f"Monitoring period is not a valid int")
		if period < 60:
			raise self.MonitoringConfiguration.InvalidMonitoringConfigurationException(f"Monitoring period cannot be smaller than 60 seconds")

		return self.MonitoringConfiguration(
			server_url=server_url,
			bot_name=bot_name,
			period=period,
			mail_addresses=mail_addresses,
			sms_numbers=sms_numbers,
			webhook_urls=webhook_urls,
			send_test_notification=send_test_notification
		)

	#####
	# Monitoring routine core
	#####
	async def _start_monitoring_process(self):
		tools_logger.debug("monitoring: starting process")
		# build request payload
		json_payload = {
			"bot_name": self._monitoring_configuration.bot_name,
			"duration_before_alert": 2 * self._monitoring_configuration.period,  # bot have two periods to
			"alerts": []
		}

		if len(self._monitoring_configuration.mail_addresses):
			json_payload["alerts"].append({"type": "mail", "recipients": self._monitoring_configuration.mail_addresses})
		if len(self._monitoring_configuration.sms_numbers):
			json_payload["alerts"].append({"type": "sms", "recipients": self._monitoring_configuration.sms_numbers})
		if len(self._monitoring_configuration.webhook_urls):
			json_payload["alerts"].append({"type": "webhook", "recipients": self._monitoring_configuration.webhook_urls})

		# register on server
		try:
			status, body = await self.async_http_request(self._monitoring_configuration.server_url + "/register", "POST", json.dumps(json_payload))
			if status != 200:
				raise RuntimeError(f"Unable to subscribe to monitoring: {status}: {body}")
			subscription_id: str = body
		except Exception as e:
			tools_logger.exception(f"{self.name}: exception during subscription process: {e} ({json_payload})")
			raise e

		# send test notification if necessary
		if self._monitoring_configuration.send_test_notification:
			try:
				status, body = await self.async_http_request(url=self._monitoring_configuration.server_url + "/test", method="POST", payload=subscription_id)
				if status != 200:
					raise RuntimeError(f"An error occurred when sent test notification: {status}: {body}")
			except Exception as e:
				tools_logger.exception(f"monitoring: exception when sending test message: {e} ({subscription_id})")
				raise e

		# add a background task to refresh subscription on monitoring server
		async def refresh_subscription_task():
			while True:
				try:
					# execute arbitrary code before refreshing subscription on server
					res = self.monitoring_routine()
					if iscoroutine(res):
						await res
					if self.argument_monitoring_routine:
						res = self.argument_monitoring_routine(self)
						if iscoroutine(res):
							await res

					# refresh subscription on server
					refresh_status, refresh_body = await self.async_http_request(url=self._monitoring_configuration.server_url + "/refresh", method="POST", payload=subscription_id)

					# subscription expired on server, launch a new subscription process in background and end this task
					if refresh_status == 404:
						tools_logger.error(f"{self.name}: refresh: server returned a 404 error, restarting monitoring process")
						self.add_background_task(self._start_monitoring_process())
						return

					if refresh_status != 200:
						raise Exception(f"Server returned an error: {refresh_status}")

					tools_logger.debug(f"{self.name}: pinged monitoring server")
					await asyncio.sleep(self._monitoring_configuration.period)

				except Exception:
					# an error occurred try to refresh faster before subscription expiration
					tools_logger.exception(f"{self.name}: an error occurred during refresh process ({subscription_id})")
					await asyncio.sleep(5)

		# start task if running, else wait for client start (avoid asyncio loop issues)
		self.add_background_task(refresh_subscription_task(), "monitoring-refresh-task")

	####
	# Http methods (we use aiohttp if available, else we run urllib in executor)
	# This returns a tuple with http response status and body
	####
	async def async_http_request(self, url: str, method: str, payload: str = None) -> tuple[int, str]:
		if "aiohttp" in globals():
			async with self.aiohttp_session.request(method=method, url=url, data=payload) as response:
				return response.status, await response.text()
		else:
			def sync_http_request() -> tuple[int, str]:
				try:
					with urllib.request.urlopen(urllib.request.Request(url=url, method=method, data=payload.encode() if payload else None)) as resp:
						return (resp.status, resp.read())
				except urllib.error.HTTPError as error:
					return (error.status, error.read().decode())
			return await asyncio.get_event_loop().run_in_executor(None, sync_http_request)

	class MonitoringConfiguration():
		def __init__(
				self,
				server_url: str,
				bot_name: str,
				period: int = 120,
				mail_addresses: list[str] = (),
				sms_numbers: list[str] = (),
				webhook_urls: list[str] = (),
				send_test_notification: bool = False

		):
			self.server_url: str = server_url
			self.bot_name: str = bot_name
			self.period: int = period
			self.mail_addresses: list[str] = mail_addresses
			self.sms_numbers: list[str] = sms_numbers
			self.webhook_urls: list[str] = webhook_urls
			self.send_test_notification: bool = send_test_notification

		class InvalidMonitoringConfigurationException(Exception):
			pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: olvid/daemon/admin/v1/identity_admin.proto
# Protobuf Python Version: 5.27.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    3,
    '',
    'olvid/daemon/admin/v1/identity_admin.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from .....olvid.daemon.datatypes.v1 import identity_pb2 as olvid_dot_daemon_dot_datatypes_dot_v1_dot_identity__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n*olvid/daemon/admin/v1/identity_admin.proto\x12\x15olvid.daemon.admin.v1\x1a(olvid/daemon/datatypes/v1/identity.proto\"h\n\x13IdentityListRequest\x12\x46\n\x06\x66ilter\x18\x01 \x01(\x0b\x32).olvid.daemon.datatypes.v1.IdentityFilterH\x00R\x06\x66ilter\x88\x01\x01\x42\t\n\x07_filter\"[\n\x14IdentityListResponse\x12\x43\n\nidentities\x18\x01 \x03(\x0b\x32#.olvid.daemon.datatypes.v1.IdentityR\nidentities\":\n\x17IdentityAdminGetRequest\x12\x1f\n\x0bidentity_id\x18\x01 \x01(\x04R\nidentityId\"[\n\x18IdentityAdminGetResponse\x12?\n\x08identity\x18\x01 \x01(\x0b\x32#.olvid.daemon.datatypes.v1.IdentityR\x08identity\"\xc8\x01\n\x12IdentityNewRequest\x12U\n\x10identity_details\x18\x01 \x01(\x0b\x32*.olvid.daemon.datatypes.v1.IdentityDetailsR\x0fidentityDetails\x12\"\n\nserver_url\x18\x02 \x01(\tH\x00R\tserverUrl\x88\x01\x01\x12\x1c\n\x07\x61pi_key\x18\x03 \x01(\tH\x01R\x06\x61piKey\x88\x01\x01\x42\r\n\x0b_server_urlB\n\n\x08_api_key\"V\n\x13IdentityNewResponse\x12?\n\x08identity\x18\x01 \x01(\x0b\x32#.olvid.daemon.datatypes.v1.IdentityR\x08identity\"8\n\x15IdentityDeleteRequest\x12\x1f\n\x0bidentity_id\x18\x01 \x01(\x04R\nidentityId\"\x18\n\x16IdentityDeleteResponse\"K\n\x1aIdentityKeycloakNewRequest\x12-\n\x12\x63onfiguration_link\x18\x01 \x01(\tR\x11\x63onfigurationLink\"^\n\x1bIdentityKeycloakNewResponse\x12?\n\x08identity\x18\x01 \x01(\x0b\x32#.olvid.daemon.datatypes.v1.IdentityR\x08identityB\xb7\x01\n\x19\x63om.olvid.daemon.admin.v1B\x12IdentityAdminProtoP\x01Z\x0folvid.io/daemon\xa2\x02\x03ODA\xaa\x02\x15Olvid.Daemon.Admin.V1\xca\x02\x15Olvid\\Daemon\\Admin\\V1\xe2\x02!Olvid\\Daemon\\Admin\\V1\\GPBMetadata\xea\x02\x18Olvid::Daemon::Admin::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'olvid.daemon.admin.v1.identity_admin_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\031com.olvid.daemon.admin.v1B\022IdentityAdminProtoP\001Z\017olvid.io/daemon\242\002\003ODA\252\002\025Olvid.Daemon.Admin.V1\312\002\025Olvid\\Daemon\\Admin\\V1\342\002!Olvid\\Daemon\\Admin\\V1\\GPBMetadata\352\002\030Olvid::Daemon::Admin::V1'
  _globals['_IDENTITYLISTREQUEST']._serialized_start=111
  _globals['_IDENTITYLISTREQUEST']._serialized_end=215
  _globals['_IDENTITYLISTRESPONSE']._serialized_start=217
  _globals['_IDENTITYLISTRESPONSE']._serialized_end=308
  _globals['_IDENTITYADMINGETREQUEST']._serialized_start=310
  _globals['_IDENTITYADMINGETREQUEST']._serialized_end=368
  _globals['_IDENTITYADMINGETRESPONSE']._serialized_start=370
  _globals['_IDENTITYADMINGETRESPONSE']._serialized_end=461
  _globals['_IDENTITYNEWREQUEST']._serialized_start=464
  _globals['_IDENTITYNEWREQUEST']._serialized_end=664
  _globals['_IDENTITYNEWRESPONSE']._serialized_start=666
  _globals['_IDENTITYNEWRESPONSE']._serialized_end=752
  _globals['_IDENTITYDELETEREQUEST']._serialized_start=754
  _globals['_IDENTITYDELETEREQUEST']._serialized_end=810
  _globals['_IDENTITYDELETERESPONSE']._serialized_start=812
  _globals['_IDENTITYDELETERESPONSE']._serialized_end=836
  _globals['_IDENTITYKEYCLOAKNEWREQUEST']._serialized_start=838
  _globals['_IDENTITYKEYCLOAKNEWREQUEST']._serialized_end=913
  _globals['_IDENTITYKEYCLOAKNEWRESPONSE']._serialized_start=915
  _globals['_IDENTITYKEYCLOAKNEWRESPONSE']._serialized_end=1009
# @@protoc_insertion_point(module_scope)



from typing import Optional

from .OlvidClient import OlvidClient
from .. import datatypes
from ..listeners.Command import Command, CommandHolder
from ..listeners.GenericNotificationListener import GenericNotificationListener
from ..listeners.Notifications import NOTIFICATIONS
from ..listeners import ListenersImplementation as listeners


class OlvidBot(OlvidClient, CommandHolder):
	"""
	OlvidBot: OlvidClient extended to add notification handlers.

	As OlvidBot extends OlvidClient it has the same constraints, for example it needs to find a valid client_key to
	connect to a daemon (see OlvidClient for more information).

	OlvidBot implements a set of method named on_something. There is one method for each gRPC notification method.
	On instantiation overwritten methods will automatically be subscribed as notification listener.
	For example:
	```
	class Bot(OlvidBot)
		async def on_message_received(self, message: datatypes.Message):
			print(message)
	```
	Every time Bot class is instantiated it will add a listener to message_received notification with the method as handler.

	OlvidBot can also add Command objects with add_command method. Command are specific listeners.
	They subclass listeners.MessageReceivedListener, and they are created with a regexp filter that will filter notifications.
	Only messages that match the regexp will raise a notification.
	Commands can be added using OlvidBot.command decorator:
	For example:
	```
	class Bot(OlvidBot)
		OlvidBot.command(regexp_filter="^!help")
		async def on_message_received(self, message: datatypes.Message):
			await message.reply("Help message")
	```
	"""
	def __init__(self, bot_name: str = None, client_key: Optional[str] = None, server_target: Optional[str] = None,
				parent_client: Optional['OlvidClient'] = None):
		OlvidClient.__init__(self, client_key=client_key, server_target=server_target, parent_client=parent_client)
		CommandHolder.__init__(self)
		self._name = bot_name if bot_name is not None else self.__class__.__name__

		# add NotificationListener overwritten methods to listeners
		listeners_to_add = self._get_listeners_to_add()
		for listener in listeners_to_add:
			self.add_listener(listener)

	# read only properties
	@property
	def name(self) -> str:
		return self._name

	#####
	# CommandHolder interface implementation
	#####
	def add_command(self, command: Command):
		self.add_listener(command)

	def remove_command(self, command: Command):
		self.remove_listener(command)

	def is_message_body_a_valid_command(self, body: str) -> bool:
		return any([isinstance(listener, Command) and listener.match_str(body) for listener in self._listeners_set])

	#####
	# Common methods
	#####
	def __str__(self) -> str:
		current_listeners = [f"{listener.listener_key}: (finished: {listener.is_finished})" for listener in self._listeners_set]
		return f"{self.name}: {', '.join(current_listeners)}"

	#####
	# tools
	#####
	def print(self, *args, **kwargs):
		print(f"{self._name}:", *args, **kwargs)

	####
	# NotificationHandler method
	####
	def _get_listeners_to_add(self) -> list[GenericNotificationListener]:
		listeners_list: list = list()
		for notification in NOTIFICATIONS:
			if not self._was_notification_listener_method_overwritten(notification):
				continue
			camel_case_notification_name = "".join(s.title() for s in notification.name.split("_"))
			listener = getattr(listeners, f"{camel_case_notification_name}Listener")(handler=getattr(self, f"on_{notification.name.lower()}"))
			listeners_list.append(listener)
		return listeners_list

	# check if a method was overwritten
	def _was_notification_listener_method_overwritten(self, notification: NOTIFICATIONS) -> bool:
		method_name = f"on_{notification.name.lower()}"
		# check method exists
		if not hasattr(self, method_name):
			return False
		# if listener method is different from original OlvidBot method, return it
		if getattr(type(self), method_name) != getattr(OlvidBot, method_name, None):
			return True
		# listener was not overwritten
		return False

	#####
	# Notification handlers: override any of this method to easily add a handler for any type of notification
	#####

	####################################################################################################################
	##### WARNING: DO NOT EDIT: this code is automatically generated, see overlay_generator/generate_olvid_client_code.py
	####################################################################################################################
		# InvitationNotificationService
	async def on_invitation_received(self, invitation: datatypes.Invitation):
		pass

	async def on_invitation_sent(self, invitation: datatypes.Invitation):
		pass

	async def on_invitation_deleted(self, invitation: datatypes.Invitation):
		pass

	async def on_invitation_updated(self, invitation: datatypes.Invitation, previous_invitation_status: datatypes.Invitation.Status):
		pass

	# ContactNotificationService
	async def on_contact_new(self, contact: datatypes.Contact):
		pass

	async def on_contact_deleted(self, contact: datatypes.Contact):
		pass

	async def on_contact_details_updated(self, contact: datatypes.Contact, previous_details: datatypes.IdentityDetails):
		pass

	# GroupNotificationService
	async def on_group_new(self, group: datatypes.Group):
		pass

	async def on_group_deleted(self, group: datatypes.Group):
		pass

	async def on_group_name_updated(self, group: datatypes.Group, previous_name: str):
		pass

	async def on_group_description_updated(self, group: datatypes.Group, previous_description: str):
		pass

	async def on_group_pending_member_added(self, group: datatypes.Group, pending_member: datatypes.PendingGroupMember):
		pass

	async def on_group_pending_member_removed(self, group: datatypes.Group, pending_member: datatypes.PendingGroupMember):
		pass

	async def on_group_member_joined(self, group: datatypes.Group, member: datatypes.GroupMember):
		pass

	async def on_group_member_left(self, group: datatypes.Group, member: datatypes.GroupMember):
		pass

	async def on_group_own_permissions_updated(self, group: datatypes.Group, permissions: datatypes.GroupMemberPermissions, previous_permissions: datatypes.GroupMemberPermissions):
		pass

	async def on_group_member_permissions_updated(self, group: datatypes.Group, member: datatypes.GroupMember, previous_permissions: datatypes.GroupMemberPermissions):
		pass

	async def on_group_update_in_progress(self, group_id: int):
		pass

	async def on_group_update_finished(self, group_id: int):
		pass

	# DiscussionNotificationService
	async def on_discussion_new(self, discussion: datatypes.Discussion):
		pass

	async def on_discussion_locked(self, discussion: datatypes.Discussion):
		pass

	async def on_discussion_title_updated(self, discussion: datatypes.Discussion, previous_title: str):
		pass

	async def on_discussion_settings_updated(self, new_settings: datatypes.DiscussionSettings, previous_settings: datatypes.DiscussionSettings):
		pass

	# MessageNotificationService
	async def on_message_received(self, message: datatypes.Message):
		pass

	async def on_message_sent(self, message: datatypes.Message):
		pass

	async def on_message_deleted(self, message: datatypes.Message):
		pass

	async def on_message_body_updated(self, message: datatypes.Message, previous_body: str):
		pass

	async def on_message_uploaded(self, message: datatypes.Message):
		pass

	async def on_message_delivered(self, message: datatypes.Message):
		pass

	async def on_message_read(self, message: datatypes.Message):
		pass

	async def on_message_location_received(self, message: datatypes.Message):
		pass

	async def on_message_location_sharing_start(self, message: datatypes.Message):
		pass

	async def on_message_location_sharing_update(self, message: datatypes.Message, previous_location: datatypes.MessageLocation):
		pass

	async def on_message_location_sharing_end(self, message: datatypes.Message):
		pass

	async def on_message_reaction_added(self, message: datatypes.Message, reaction: datatypes.MessageReaction):
		pass

	async def on_message_reaction_updated(self, message: datatypes.Message, reaction: datatypes.MessageReaction, previous_reaction: datatypes.MessageReaction):
		pass

	async def on_message_reaction_removed(self, message: datatypes.Message, reaction: datatypes.MessageReaction):
		pass

	# AttachmentNotificationService
	async def on_attachment_received(self, attachment: datatypes.Attachment):
		pass

	async def on_attachment_uploaded(self, attachment: datatypes.Attachment):
		pass



from typing import Callable

import grpc
from olvid import OlvidAdminClient, OlvidClient, datatypes, listeners


async def listen(identity_id: int = 0, quiet: bool = False, notifications_to_listen: str = ""):
	admin_client: OlvidAdminClient = OlvidAdminClient(identity_id=0)
	# create clients, one per identity
	clients: list[OlvidClient] = []
	try:
		if identity_id <= 0:
			async for identity in admin_client.admin_identity_list():
				clients.append(OlvidAdminClient(identity_id=identity.id))
		else:
			clients.append(OlvidAdminClient(identity_id=identity_id))
	except grpc.aio.AioRpcError as e:
		print(e.details())
		return

	# determine notifications to listen to
	if not notifications_to_listen:
		notifications: list[listeners.NOTIFICATIONS] = [notification for notification in listeners.NOTIFICATIONS]
	else:
		notifications: list[listeners.NOTIFICATIONS] = [getattr(listeners.NOTIFICATIONS, notif_name) for notif_name in notifications_to_listen.split(",")]

	# add listeners
	for client in clients:
		identity: datatypes.Identity = await client.identity_get()
		for notification in notifications:
			listener_class_name = f"{''.join(s.title() for s in notification.name.split('_'))}Listener"
			listener_class = getattr(listeners, listener_class_name)
			client.add_listener(listener_class(handler=await get_notification_handler(identity, notification, quiet)))

	for client in clients:
		await client.wait_for_listeners_end()


async def get_notification_handler(identity: datatypes.Identity, notification_type: listeners.NOTIFICATIONS, quiet: bool) -> Callable:
	def notification_handler(*fields):
		if quiet:
			print(f"{identity.id:2}: {notification_type.name}")
		else:
			print(f"{identity.id:2}: {notification_type.name:20}: {', '.join([field_to_str(field) for field in fields])}")
	return notification_handler


def field_to_str(field) -> str:
	if isinstance(field, datatypes.Message):
		id_str = f"{'O' if field.id.type == datatypes.MessageId.Type.TYPE_OUTBOUND else 'I'}{field.id.id}"
		return f"({id_str}) {field.body}"
	elif isinstance(field, datatypes.Attachment):
		id_str = f"{'O' if field.id.type == datatypes.AttachmentId.Type.TYPE_OUTBOUND else 'I'}{field.id.id})"
		return f"({id_str}) {field.file_name}"
	elif isinstance(field, datatypes.Group):
		return f"({field.id}) {field.name}"
	elif isinstance(field, datatypes.Discussion):
		return f"({field.id}) {field.title}"
	elif isinstance(field, datatypes.Contact):
		return f"({field.id}) {field.display_name}"
	elif isinstance(field, datatypes.Invitation):
		return f"({field.id}) {field.display_name} {field.status.name}"
	else:
		return f"{field}"



from .admin import ClientKeyListRequest, ClientKeyListResponse, ClientKeyGetRequest, ClientKeyGetResponse, ClientKeyNewRequest, ClientKeyNewResponse, ClientKeyDeleteRequest, ClientKeyDeleteResponse, IdentityListRequest, IdentityListResponse, IdentityAdminGetRequest, IdentityAdminGetResponse, IdentityNewRequest, IdentityNewResponse, IdentityDeleteRequest, IdentityDeleteResponse, IdentityKeycloakNewRequest, IdentityKeycloakNewResponse, ClientKeyAdminServiceStub, IdentityAdminServiceStub
del admin



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: olvid/daemon/notification/v1/invitation_notifications.proto
# Protobuf Python Version: 5.27.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    3,
    '',
    'olvid/daemon/notification/v1/invitation_notifications.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from .....olvid.daemon.datatypes.v1 import invitation_pb2 as olvid_dot_daemon_dot_datatypes_dot_v1_dot_invitation__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n;olvid/daemon/notification/v1/invitation_notifications.proto\x12\x1colvid.daemon.notification.v1\x1a*olvid/daemon/datatypes/v1/invitation.proto\"+\n)SubscribeToInvitationReceivedNotification\"g\n\x1eInvitationReceivedNotification\x12\x45\n\ninvitation\x18\x01 \x01(\x0b\x32%.olvid.daemon.datatypes.v1.InvitationR\ninvitation\"\'\n%SubscribeToInvitationSentNotification\"c\n\x1aInvitationSentNotification\x12\x45\n\ninvitation\x18\x01 \x01(\x0b\x32%.olvid.daemon.datatypes.v1.InvitationR\ninvitation\"*\n(SubscribeToInvitationDeletedNotification\"f\n\x1dInvitationDeletedNotification\x12\x45\n\ninvitation\x18\x01 \x01(\x0b\x32%.olvid.daemon.datatypes.v1.InvitationR\ninvitation\"*\n(SubscribeToInvitationUpdatedNotification\"\xd2\x01\n\x1dInvitationUpdatedNotification\x12\x45\n\ninvitation\x18\x01 \x01(\x0b\x32%.olvid.daemon.datatypes.v1.InvitationR\ninvitation\x12j\n\x1aprevious_invitation_status\x18\x02 \x01(\x0e\x32,.olvid.daemon.datatypes.v1.Invitation.StatusR\x18previousInvitationStatusB\xe4\x01\n com.olvid.daemon.notification.v1B\x1cInvitationNotificationsProtoP\x01Z\x0folvid.io/daemon\xa2\x02\x03ODN\xaa\x02\x1cOlvid.Daemon.Notification.V1\xca\x02\x1cOlvid\\Daemon\\Notification\\V1\xe2\x02(Olvid\\Daemon\\Notification\\V1\\GPBMetadata\xea\x02\x1fOlvid::Daemon::Notification::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'olvid.daemon.notification.v1.invitation_notifications_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n com.olvid.daemon.notification.v1B\034InvitationNotificationsProtoP\001Z\017olvid.io/daemon\242\002\003ODN\252\002\034Olvid.Daemon.Notification.V1\312\002\034Olvid\\Daemon\\Notification\\V1\342\002(Olvid\\Daemon\\Notification\\V1\\GPBMetadata\352\002\037Olvid::Daemon::Notification::V1'
  _globals['_SUBSCRIBETOINVITATIONRECEIVEDNOTIFICATION']._serialized_start=137
  _globals['_SUBSCRIBETOINVITATIONRECEIVEDNOTIFICATION']._serialized_end=180
  _globals['_INVITATIONRECEIVEDNOTIFICATION']._serialized_start=182
  _globals['_INVITATIONRECEIVEDNOTIFICATION']._serialized_end=285
  _globals['_SUBSCRIBETOINVITATIONSENTNOTIFICATION']._serialized_start=287
  _globals['_SUBSCRIBETOINVITATIONSENTNOTIFICATION']._serialized_end=326
  _globals['_INVITATIONSENTNOTIFICATION']._serialized_start=328
  _globals['_INVITATIONSENTNOTIFICATION']._serialized_end=427
  _globals['_SUBSCRIBETOINVITATIONDELETEDNOTIFICATION']._serialized_start=429
  _globals['_SUBSCRIBETOINVITATIONDELETEDNOTIFICATION']._serialized_end=471
  _globals['_INVITATIONDELETEDNOTIFICATION']._serialized_start=473
  _globals['_INVITATIONDELETEDNOTIFICATION']._serialized_end=575
  _globals['_SUBSCRIBETOINVITATIONUPDATEDNOTIFICATION']._serialized_start=577
  _globals['_SUBSCRIBETOINVITATIONUPDATEDNOTIFICATION']._serialized_end=619
  _globals['_INVITATIONUPDATEDNOTIFICATION']._serialized_start=622
  _globals['_INVITATIONUPDATEDNOTIFICATION']._serialized_end=832
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: olvid/daemon/datatypes/v1/invitation.proto
# Protobuf Python Version: 5.27.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    3,
    '',
    'olvid/daemon/datatypes/v1/invitation.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n*olvid/daemon/datatypes/v1/invitation.proto\x12\x19olvid.daemon.datatypes.v1\"\xca\x05\n\nInvitation\x12\x0e\n\x02id\x18\x01 \x01(\x04R\x02id\x12\x44\n\x06status\x18\x02 \x01(\x0e\x32,.olvid.daemon.datatypes.v1.Invitation.StatusR\x06status\x12!\n\x0c\x64isplay_name\x18\x03 \x01(\tR\x0b\x64isplayName\x12\x1c\n\ttimestamp\x18\x04 \x01(\x04R\ttimestamp\x12\x10\n\x03sas\x18\x05 \x01(\tR\x03sas\"\x92\x04\n\x06Status\x12\x16\n\x12STATUS_UNSPECIFIED\x10\x00\x12(\n$STATUS_INVITATION_WAIT_YOU_TO_ACCEPT\x10\x01\x12\'\n#STATUS_INVITATION_WAIT_IT_TO_ACCEPT\x10\x02\x12(\n$STATUS_INVITATION_STATUS_IN_PROGRESS\x10\x03\x12/\n+STATUS_INVITATION_WAIT_YOU_FOR_SAS_EXCHANGE\x10\x04\x12.\n*STATUS_INVITATION_WAIT_IT_FOR_SAS_EXCHANGE\x10\x05\x12)\n%STATUS_INTRODUCTION_WAIT_IT_TO_ACCEPT\x10\x07\x12*\n&STATUS_INTRODUCTION_WAIT_YOU_TO_ACCEPT\x10\x08\x12\x32\n.STATUS_ONE_TO_ONE_INVITATION_WAIT_IT_TO_ACCEPT\x10\t\x12\x33\n/STATUS_ONE_TO_ONE_INVITATION_WAIT_YOU_TO_ACCEPT\x10\n\x12.\n*STATUS_GROUP_INVITATION_WAIT_YOU_TO_ACCEPT\x10\x0b\x12\"\n\x1eSTATUS_GROUP_INVITATION_FROZEN\x10\x0c\"\xf0\x03\n\x10InvitationFilter\x12I\n\x06status\x18\x01 \x01(\x0e\x32,.olvid.daemon.datatypes.v1.Invitation.StatusH\x00R\x06status\x88\x01\x01\x12I\n\x04type\x18\x02 \x01(\x0e\x32\x30.olvid.daemon.datatypes.v1.InvitationFilter.TypeH\x01R\x04type\x88\x01\x01\x12\x33\n\x13\x64isplay_name_search\x18\x03 \x01(\tH\x02R\x11\x64isplayNameSearch\x88\x01\x01\x12(\n\rmin_timestamp\x18\x04 \x01(\x04H\x03R\x0cminTimestamp\x88\x01\x01\x12(\n\rmax_timestamp\x18\x05 \x01(\x04H\x04R\x0cmaxTimestamp\x88\x01\x01\"m\n\x04Type\x12\x14\n\x10TYPE_UNSPECIFIED\x10\x00\x12\x13\n\x0fTYPE_INVITATION\x10\x01\x12\x15\n\x11TYPE_INTRODUCTION\x10\x02\x12\x0e\n\nTYPE_GROUP\x10\x03\x12\x13\n\x0fTYPE_ONE_TO_ONE\x10\x04\x42\t\n\x07_statusB\x07\n\x05_typeB\x16\n\x14_display_name_searchB\x10\n\x0e_min_timestampB\x10\n\x0e_max_timestampB\xc8\x01\n\x1d\x63om.olvid.daemon.datatypes.v1B\x0fInvitationProtoP\x01Z\x0folvid.io/daemon\xa2\x02\x03ODD\xaa\x02\x19Olvid.Daemon.Datatypes.V1\xca\x02\x19Olvid\\Daemon\\Datatypes\\V1\xe2\x02%Olvid\\Daemon\\Datatypes\\V1\\GPBMetadata\xea\x02\x1cOlvid::Daemon::Datatypes::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'olvid.daemon.datatypes.v1.invitation_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\035com.olvid.daemon.datatypes.v1B\017InvitationProtoP\001Z\017olvid.io/daemon\242\002\003ODD\252\002\031Olvid.Daemon.Datatypes.V1\312\002\031Olvid\\Daemon\\Datatypes\\V1\342\002%Olvid\\Daemon\\Datatypes\\V1\\GPBMetadata\352\002\034Olvid::Daemon::Datatypes::V1'
  _globals['_INVITATION']._serialized_start=74
  _globals['_INVITATION']._serialized_end=788
  _globals['_INVITATION_STATUS']._serialized_start=258
  _globals['_INVITATION_STATUS']._serialized_end=788
  _globals['_INVITATIONFILTER']._serialized_start=791
  _globals['_INVITATIONFILTER']._serialized_end=1287
  _globals['_INVITATIONFILTER_TYPE']._serialized_start=1098
  _globals['_INVITATIONFILTER_TYPE']._serialized_end=1207
# @@protoc_insertion_point(module_scope)
